"""
hdl21 ProtoBuf Import 
"""
from types import SimpleNamespace
from typing import Union, Any, Dict, List, Optional

# Local imports
# Proto-definitions
import vlsir
import vlsir.circuit_pb2 as vckt

# HDL
from ..prefix import Prefix, Prefixed
from ..module import Module
from ..external_module import ExternalModule
from ..instance import Instance
from ..signal import Signal, PortDir, Visibility
from ..slice import Slice
from ..concat import Concat
from ..literal import Literal
from .. import primitives
from ..primitives import Primitive, Vpulse


def from_proto(pkg: vckt.Package) -> SimpleNamespace:
    """Convert Proto-defined Package `pkg` to a namespace-full of Modules."""
    importer = ProtoImporter(pkg)
    return importer.import_()


class ProtoImporter:
    """Protobuf Package Importer.
    Collects all `Modules` defined in Protobuf-sourced primary-argument `pkg` into a Python `types.SimpleNamespace`.
    """

    def __init__(self, pkg: vckt.Package):
        self.pkg = pkg
        self.modules = dict()  # Dict of names to Modules
        self.ext_modules = dict()  # Dict of qual-names to ExternalModules
        self.ns = SimpleNamespace()
        self.ns.name = pkg.domain

    def import_(self) -> SimpleNamespace:
        """Import the top-level `Package` to a Python namespace"""
        # Walk through each proto-defined Module
        # External modules first, as we know they have no dependencies
        for emod in self.pkg.ext_modules:
            self.import_external_module(emod)
        for pmod in self.pkg.modules:
            self.import_module(pmod)
        # Return our collected namespace
        return self.ns

    def get_namespace(self, path: List[str]) -> SimpleNamespace:
        """Get a (potentially nested) namespace at `path`, creating levels along the way if necessary ."""
        ns = self.ns
        for part in path:
            attr = getattr(ns, part, None)
            if attr is None:  # Create a new Namespce
                new_ns = SimpleNamespace()
                new_ns.name = part
                setattr(ns, part, new_ns)
                ns = new_ns
            elif isinstance(attr, SimpleNamespace):
                ns = attr
            else:
                raise RuntimeError(f"Invalid namespace path {path} overwriting {attr}")
        return ns

    def import_external_module(self, pmod: vckt.ExternalModule) -> ExternalModule:
        """Convert Proto-Module `emod` to an `hdl21.ExternalModule`"""

        # Check our cache for whether a module by the same name has been imported
        key = (pmod.name.domain, pmod.name.name)
        if key in self.ext_modules:
            conflict = self.ext_modules[key]
            msg = f"Cannot import conflicting definitions of {pmod} and {conflict}"
            raise RuntimeError(msg)

        # Import the port list
        port_list = import_ports_and_signals(pmod)

        # Create the ExternalModule
        emod = ExternalModule(
            name=pmod.name.name,
            domain=pmod.name.domain,
            desc=pmod.desc,
            port_list=port_list,
            paramtype=dict,  # FIXME: should these be stored in the serialization schema?
        )
        # Give it a (non-initializer) value for its `importpath`
        emod._importpath = [pmod.name.domain]
        # Cache and return it
        self.ext_modules[key] = emod
        return emod

    def import_module(self, pmod: vckt.Module) -> Module:
        """Convert Proto-Module `pmod` to an `hdl21.Module`"""
        if pmod.name in self.modules:
            raise RuntimeError(f"Proto Import Error: Redefined Module {pmod.name}")

        # Create the Module
        module = Module()
        # Get or create its namespace, and set its name
        path = pmod.name.split(".")  # Path-parts are dot-separated
        ns = self.get_namespace(path[:-1])
        # Save the import-path and name
        module._importpath = path[:-1]
        module.name = path[-1]

        # Import its signals and ports
        for sig in import_ports_and_signals(pmod):
            module.add(sig)

        # Lap through instances & connect them
        for pinst in pmod.instances:
            inst = self.import_instance(pinst)
            module.add(inst)

            # Make the instance's connections
            for pconn in pinst.connections:
                if pconn.portname not in inst.of.ports:
                    msg = f"Invalid Port {pconn.portname} on {inst} in Module {module.name}"
                    raise RuntimeError(msg)
                # Import the Signal-object
                conn = import_connection_target(pconn.target, module)
                # And connect it to the Instance
                inst.connect(pconn.portname, conn)

        # Import any literal content
        for plit in pmod.literals:
            module.literals.append(Literal(text=plit))

        # Add the Module to our cache and return-namespace, and return it
        self.modules[pmod.name] = module
        setattr(ns, module.name, module)
        return module

    def import_instance(self, pinst: vckt.Instance) -> Instance:
        """Convert Proto-Instance `pinst` to an `hdl21.Instance`.
        Requires an available Module-definition to be referenced.
        Connections are *not* performed inside this method."""

        # Also a small piece of proof that Google hates Python.
        ref = pinst.module
        if ref.WhichOneof("to") == "local":
            # Internally-defined Module
            module = self.modules.get(ref.local, None)
            if module is None:
                raise RuntimeError(f"Invalid undefined Module {ref.local} ")
            if len(pinst.parameters):
                msg = f"Invalid Instance {pinst} with of Module {module} - does not accept Parameters"
                raise RuntimeError(msg)
            target = module

        elif ref.WhichOneof("to") == "external":  # Defined outside package
            # Import all of its instance parameters to a dict
            params = import_parameters(pinst.parameters)

            # First check the priviledged/ internally-defined domains
            if ref.external.domain == "vlsir.primitives":
                # Import a VLSIR primitive to an ideal element, and convert its parameters
                target = import_vlsir_primitive(ref.external)
                remapped_params = import_primitive_params(target, params)
                params = target.Params(**remapped_params)

            elif ref.external.domain in (
                "hdl21.primitives",
                "hdl21.ideal",
            ):
                # Retrieve the Primitive from `hdl21.primitives`, and convert its parameters
                target = import_hdl21_primitive(ref.external)
                params = target.Params(**params)

            else:  # Externally-defined `ExternalModule`
                # These must be declared in our `Package` being imported. Look up its header-info from `ext_modules`.
                key = (ref.external.domain, ref.external.name)
                target = self.ext_modules.get(key, None)
                if target is None:
                    msg = f"Invalid Instance of undefined External Module {key}"
                    raise RuntimeError(msg)

            # Call the `target` ExternalModule/ Primitive it with the parameters, making an instantiable `Call`-object.
            target = target(params)

        else:
            raise ValueError

        return Instance(name=pinst.name, of=target)


def import_ports_and_signals(
    pmod: Union[vckt.Module, vckt.ExternalModule]
) -> List[Signal]:
    """Import all Ports and Signals from Proto-Module `pmod`.
    Returns them as a single list, which can serve as arguments to `Module.add` or `ExternalModule.port_list`.
    """

    # Keep a dictionary from name: Signal imported
    signals: Dict[str, Signal] = {}
    for psig in pmod.signals:
        signals[psig.name] = Signal(name=psig.name, width=psig.width)

    # Convert the entries of `signals` that are ports
    for pport in pmod.ports:
        dir_ = import_port_dir(pport)
        if pport.signal not in signals:
            msg = f"Port {pport} missing Signal in {signals}"
            raise RuntimeError(msg)
        signals[pport.signal].direction = dir_
        signals[pport.signal].vis = Visibility.PORT

    return list(signals.values())


def import_hdl21_primitive(pref: vlsir.utils.QualifiedName) -> Primitive:
    """Convert an `hdl21.primitives` or `hdl21.ideal` qualified name to a Primitive."""
    if pref.domain not in ["hdl21.primitives", "hdl21.ideal"]:
        raise ValueError(f"Invalid Primitive Domain: {pref.domain}")

    # Get the Primitive from `hdl21.primitives`, or fail
    prim = getattr(primitives, pref.name, None)
    if not isinstance(prim, Primitive):
        msg = f"Attempt to import invalid `hdl21.primitive` {pref.external.name}"
        raise RuntimeError(msg)
    return prim


def import_vlsir_primitive(pref: vlsir.utils.QualifiedName) -> Primitive:
    """Import a VLSIR-defined Primitive"""
    if pref.domain != "vlsir.primitives":
        raise ValueError(f"Invalid Primitive Domain: {pref.domain}")

    # Mapping from `vlsir.primitives` to Hdl21's ideal elements
    # FIXME: specialized importing of their parameters!
    prim_map = {
        "vdc": "DcVoltageSource",
        "vpulse": "PulseVoltageSource",
        "vsin": "SineVoltageSource",
        "isource": "CurrentSource",
        "resistor": "IdealResistor",
        "capacitor": "IdealCapacitor",
        "inductor": "IdealInductor",
        "vcvs": "VoltageControlledVoltageSource",
        "vccs": "VoltageControlledCurrentSource",
        "ccvs": "CurrentControlledVoltageSource",
        "cccs": "CurrentControlledCurrentSource",
    }
    if pref.name not in prim_map:
        msg = f"Invalid or unsupported VLSIR Primitive: {pref.name}"
        raise RuntimeError(msg)

    # Get the Hdl21 Primitive class
    prim = getattr(primitives, prim_map[pref.name], None)
    if not isinstance(prim, Primitive):
        msg = f"Attempt to import invalid `hdl21.primitive` {pref.external.name}"
        raise RuntimeError(msg)
    return prim


def import_parameters(pparams: List[vlsir.Param]) -> Dict[str, Any]:
    """Import a list of Vlsir parameters to a Hdl21-style {name: value} dict."""
    return {pparam.name: import_parameter_value(pparam.value) for pparam in pparams}


def import_parameter_value(
    pparam: vlsir.ParamValue,
) -> Union[int, float, str, Prefixed]:
    """Import a `ParamValue`"""
    ptype = pparam.WhichOneof("value")
    if ptype == "int64_value":
        return int(pparam.int64_value)
    if ptype == "double_value":
        return float(pparam.double_value)
    if ptype == "string_value":
        return str(pparam.string_value)
    if ptype == "literal":
        return str(pparam.literal)
    if ptype == "prefixed":
        return import_prefixed(pparam.prefixed)
    raise ValueError(f"Invalid Parameter Type: `{ptype}`")


def import_connection_target(
    pconn: vckt.ConnectionTarget, module: Module
) -> Union[Signal, Slice, Concat]:
    """Import a Proto-defined `ConnectionTarget` into a Signal, Slice, or Concatenation"""
    # Connections are a proto `oneof` union; figure out which to import
    stype = pconn.WhichOneof("stype")
    # Concatenations are more complicated and need their own method
    if stype == "concat":
        return import_concat(pconn.concat, module)
    # For signals & slices, first sort out the signal-name, so we can grab the object from `module.namespace`
    if stype == "sig":
        sname = pconn.sig
    elif stype == "slice":
        sname = pconn.slice.signal
    else:
        raise ValueError(f"Invalid Connection Type: {pconn}")
    # Grab this Signal, if it exists
    sig = module.namespace.get(sname, None)
    if sig is None:
        # This block has held, at some points in code-history,
        # the SPICE-style "create nets from thin air" behavior.
        # That's outta here; undeclared signals produce errors instead.
        raise RuntimeError(f"Invalid Signal {sname} in Module {module.name}")
    # Now chop this up if it's a Slice
    if stype == "slice":
        start = pconn.slice.bot
        stop = pconn.slice.top + 1  # Move to Python-style exclusive indexing
        sig = Slice(parent=sig, index=slice(start, stop))
    return sig


def import_concat(pconc: vckt.Concat, module: Module) -> Concat:
    """Import a (potentially nested) Concatenation"""
    parts = []
    for ppart in pconc.parts:
        part = import_connection_target(ppart, module)
        parts.append(part)
    return Concat(*parts)


def import_port_dir(pport: vckt.Port) -> PortDir:
    """Convert between Port-Direction Enumerations"""
    if pport.direction == vckt.Port.Direction.INPUT:
        return PortDir.INPUT
    if pport.direction == vckt.Port.Direction.OUTPUT:
        return PortDir.OUTPUT
    if pport.direction == vckt.Port.Direction.INOUT:
        return PortDir.INOUT
    if pport.direction == vckt.Port.Direction.NONE:
        return PortDir.NONE
    raise ValueError


def import_prefix(vpre: vlsir.SIPrefix) -> Prefix:
    """Import an enumerated `Prefix`"""
    map = {
        vlsir.SIPrefix.YOCTO: Prefix.YOCTO,
        vlsir.SIPrefix.ZEPTO: Prefix.ZEPTO,
        vlsir.SIPrefix.ATTO: Prefix.ATTO,
        vlsir.SIPrefix.FEMTO: Prefix.FEMTO,
        vlsir.SIPrefix.PICO: Prefix.PICO,
        vlsir.SIPrefix.NANO: Prefix.NANO,
        vlsir.SIPrefix.MICRO: Prefix.MICRO,
        vlsir.SIPrefix.MILLI: Prefix.MILLI,
        vlsir.SIPrefix.CENTI: Prefix.CENTI,
        vlsir.SIPrefix.DECI: Prefix.DECI,
        vlsir.SIPrefix.DECA: Prefix.DECA,
        vlsir.SIPrefix.HECTO: Prefix.HECTO,
        vlsir.SIPrefix.KILO: Prefix.KILO,
        vlsir.SIPrefix.MEGA: Prefix.MEGA,
        vlsir.SIPrefix.GIGA: Prefix.GIGA,
        vlsir.SIPrefix.TERA: Prefix.TERA,
        vlsir.SIPrefix.PETA: Prefix.PETA,
        vlsir.SIPrefix.EXA: Prefix.EXA,
        vlsir.SIPrefix.ZETTA: Prefix.ZETTA,
        vlsir.SIPrefix.YOTTA: Prefix.YOTTA,
        vlsir.SIPrefix.UNIT: Prefix.UNIT,  # Welcome to the party, as of version 2.0!
    }
    if vpre not in map:
        raise ValueError(f"Invalid Prefix {vpre}")
    return map[vpre]


def import_prefixed(vpref: vlsir.Prefixed) -> Prefixed:
    """Import a `Prefixed` number"""

    # Import the metric prefix
    prefix = import_prefix(vpref.prefix)

    # And import the numeric part, dispatched across its type.
    ptype = vpref.WhichOneof("number")
    if ptype == "int64_value":
        number = vpref.int64_value
    elif ptype == "double_value":
        number = vpref.double_value
    elif ptype == "string_value":
        number = vpref.string_value
    else:
        raise ValueError(f"Invalid Parameter Type: `{ptype}`")

    return Prefixed(number=number, prefix=prefix)


def import_primitive_params(
    target: Primitive, params: Any
) -> Dict[str, Optional[Prefixed]]:
    """Convert the parameters of an `IDEAL` VLSIR element into a primtive form.
    Returns the result as a dictionary of {name: value}s."""

    if target is Vpulse:
        return dict(
            v1=params["v1"],
            v2=params["v2"],
            delay=params["td"],
            rise=params["tr"],
            fall=params["tf"],
            width=params["tpw"],
            period=params["tper"],
        )

    return params



""" 
# Hdl21 Unit Tests

* Exporting to VLSIR 
  * Ultimately to EDA formats 
* Importing back from VLSIR
"""

import pytest
from io import StringIO
from types import SimpleNamespace
from copy import deepcopy

# Import the PUT (package under test)
import hdl21 as h
import vlsir


@pytest.mark.xfail(reason="#1 https://github.com/dan-fritchman/Hdl21/issues/1")
def test_export_strides():
    """Test exporting connections with non-unit Slice-strides"""

    c = h.Module(name="c")
    c.p = h.Input(width=2)

    p = h.Module(name="p")
    p.s = h.Signal(width=20)
    p.c = c(p=p.s[::10])  # Connect two of the 20 bits, with stride 10

    h.netlist(h.to_proto(p), StringIO(), fmt="verilog")


def test_prim_proto1():
    # Test round-tripping primitives through proto

    @h.module
    class HasPrims:
        p = h.Signal()
        n = h.Signal()

        # Wire up a bunch of two-terminal primitives in parallel
        _rp = h.R.Params(r=50)
        r = h.Resistor(_rp)(p=p, n=n)
        _cp = h.C.Params(c=1e-12)
        c = h.Capacitor(_cp)(p=p, n=n)
        _lp = h.L.Params(l=1e-15)
        l = h.Inductor(_lp)(p=p, n=n)
        _dp = h.D.Params()
        d = h.Diode(_dp)(p=p, n=n)
        _sp = h.Short.Params()
        s = h.Short(_sp)(p=p, n=n)

    ppkg = h.to_proto(HasPrims)

    assert isinstance(ppkg, vlsir.circuit.Package)
    assert len(ppkg.modules) == 1
    assert ppkg.domain == ""

    # Check the proto-Module
    pm = ppkg.modules[0]
    assert isinstance(pm, vlsir.circuit.Module)
    assert pm.name == "hdl21.tests.test_exports.HasPrims"
    assert len(pm.ports) == 0
    assert len(pm.signals) == 2
    assert len(pm.instances) == 5
    assert len(pm.parameters) == 0
    for inst in pm.instances:
        assert isinstance(inst, vlsir.circuit.Instance)
        assert inst.module.WhichOneof("to") == "external"
        assert inst.module.external.domain in ["vlsir.primitives", "hdl21.primitives"]

    ns = h.from_proto(ppkg)

    assert isinstance(ns, SimpleNamespace)
    ns = ns.hdl21.tests.test_exports
    assert isinstance(ns, SimpleNamespace)
    assert hasattr(ns, "HasPrims")
    HasPrims = ns.HasPrims
    assert isinstance(HasPrims, h.Module)
    assert len(HasPrims.ports) == 0
    assert len(HasPrims.signals) == 2
    assert len(HasPrims.instances) == 5
    for inst in HasPrims.instances.values():
        assert isinstance(inst.of, h.primitives.PrimitiveCall)


def test_ideal_primitives():
    # Test round-tripping ideal primitives

    @h.module
    class HasPrims:
        p = h.Signal()
        n = h.Signal()
        p2 = h.Signal()
        n2 = h.Signal()

        # Wire up a bunch of two-terminal primitives in parallel
        _rp = h.R.Params(r=50)
        r = h.Resistor(_rp)(p=p, n=n)
        _cp = h.C.Params(c=1e-12)
        c = h.Capacitor(_cp)(p=p, n=n)
        _lp = h.L.Params(l=1e-15)
        l = h.Inductor(_lp)(p=p, n=n)
        _dp = h.D.Params()
        d = h.Diode(_dp)(p=p, n=n)
        _sp = h.Short.Params()
        s = h.Short(_sp)(p=p, n=n)
        _vdc = h.Vdc.Params(dc=0 * h.prefix.m, ac=0 * h.prefix.m)
        vdc = h.Vdc(_vdc)(p=p, n=n)
        _vpu = h.Vpulse.Params(
            delay=0 * h.prefix.m,
            v1=0 * h.prefix.m,
            v2=0 * h.prefix.m,
            period=0 * h.prefix.m,
            rise=0 * h.prefix.m,
            fall=0 * h.prefix.m,
            width=0 * h.prefix.m,
        )
        vpu = h.Vpulse(_vpu)(p=p, n=n)
        _vsin = h.Vsin.Params(
            voff=0 * h.prefix.m,
            vamp=0 * h.prefix.m,
            freq=0 * h.prefix.m,
            td=0 * h.prefix.m,
            phase=0 * h.prefix.m,
        )
        vsin = h.Vsin(_vsin)(p=p, n=n)
        _idc = h.Idc.Params(dc=0 * h.prefix.m)
        idc = h.Idc(_idc)(p=p, n=n)
        _vcvs = h.Vcvs.Params(gain=1 * h.prefix.m)
        vcvs = h.Vcvs(_vcvs)(p=p, n=n, cp=p2, cn=n2)
        _ccvs = h.Ccvs.Params(gain=1 * h.prefix.m)
        ccvs = h.Ccvs(_ccvs)(p=p, n=n, cp=p2, cn=n2)
        _vccs = h.Vccs.Params(gain=1 * h.prefix.m)
        vccs = h.Vccs(_vccs)(p=p, n=n, cp=p2, cn=n2)
        _cccs = h.Cccs.Params(gain=1 * h.prefix.m)
        cccs = h.Cccs(_cccs)(p=p, n=n, cp=p2, cn=n2)

    ppkg = h.to_proto(HasPrims)
    ns = h.from_proto(ppkg)


def test_proto1():
    # First Proto-export test

    m = h.Module(name="TestProto1")
    ppkg = h.to_proto(m)
    assert isinstance(ppkg, vlsir.circuit.Package)
    assert len(ppkg.modules) == 1
    assert ppkg.domain == ""
    pm = ppkg.modules[0]
    assert isinstance(pm, vlsir.circuit.Module)
    assert pm.name == "hdl21.tests.test_exports.TestProto1"
    assert len(pm.ports) == 0
    assert len(pm.instances) == 0
    assert len(pm.parameters) == 0


def test_proto2():
    # Proto-export test with some hierarchy

    Child1 = h.Module(name="Child1")
    Child1.inp = h.Input(width=8)
    Child1.out = h.Output()
    Child2 = h.Module(name="Child2")
    Child2.inp = h.Input()
    Child2.out = h.Output(width=8)
    TestProto2 = h.Module(name="TestProto2")
    TestProto2.c1 = Child1()
    TestProto2.c2 = Child2()
    TestProto2.c2(inp=TestProto2.c1.out, out=TestProto2.c1.inp)

    ppkg = h.to_proto(TestProto2)

    assert isinstance(ppkg, vlsir.circuit.Package)
    assert len(ppkg.modules) == 3
    assert ppkg.domain == ""

    # Check the first Module in, Child1
    pm = ppkg.modules[0]
    assert isinstance(pm, vlsir.circuit.Module)
    assert pm.name == "hdl21.tests.test_exports.Child1"
    assert len(pm.ports) == 2
    assert len(pm.signals) == 2
    assert len(pm.instances) == 0
    assert len(pm.parameters) == 0

    # Check the second Module in, Child2
    pm = ppkg.modules[1]
    assert isinstance(pm, vlsir.circuit.Module)
    assert pm.name == "hdl21.tests.test_exports.Child2"
    assert len(pm.ports) == 2
    assert len(pm.signals) == 2
    assert len(pm.instances) == 0
    assert len(pm.parameters) == 0

    # And check the parent module
    pm = ppkg.modules[2]
    assert isinstance(pm, vlsir.circuit.Module)
    assert pm.name == "hdl21.tests.test_exports.TestProto2"
    assert len(pm.ports) == 0
    assert len(pm.signals) == 2
    assert len(pm.instances) == 2
    assert len(pm.parameters) == 0


def test_proto3():
    # Proto-export test with some slicing and concatenation

    M1 = h.Module(name="M1")
    M1.p8 = h.Inout(width=8)
    M1.p1 = h.Inout(width=1)

    M2 = h.Module(name="M2")
    M2.s = h.Signal(width=4)
    M2.i = M1()
    M2.i(p1=M2.s[0])
    M2.i(p8=h.Concat(M2.s, h.Concat(M2.s[0], M2.s[1]), M2.s[2:4]))

    ppkg = h.to_proto(M2, domain="test_proto3")

    assert isinstance(ppkg, vlsir.circuit.Package)
    assert len(ppkg.modules) == 2
    assert ppkg.domain == "test_proto3"

    # Check the child module
    pm = ppkg.modules[0]
    assert isinstance(pm, vlsir.circuit.Module)
    assert pm.name == "hdl21.tests.test_exports.M1"
    assert len(pm.ports) == 2
    assert len(pm.signals) == 2
    assert len(pm.instances) == 0
    assert len(pm.parameters) == 0

    # And check the parent module
    pm = ppkg.modules[1]
    assert isinstance(pm, vlsir.circuit.Module)
    assert pm.name == "hdl21.tests.test_exports.M2"
    assert len(pm.ports) == 0
    assert len(pm.signals) == 1
    assert len(pm.instances) == 1
    assert len(pm.parameters) == 0

    ns = h.from_proto(ppkg)

    assert isinstance(ns, SimpleNamespace)
    ns = ns.hdl21.tests.test_exports
    assert isinstance(ns, SimpleNamespace)
    assert isinstance(ns.M1, h.Module)
    assert len(ns.M1.ports) == 2
    assert len(ns.M1.signals) == 0
    assert len(ns.M1.instances) == 0

    assert isinstance(M2, h.Module)
    assert len(M2.ports) == 0
    assert len(M2.signals) == 1
    assert len(M2.instances) == 1
    assert "s" in M2.signals
    assert "i" in M2.instances
    inst = M2.i
    assert isinstance(inst.conns["p1"], h.Slice)
    assert inst.conns["p1"].parent is M2.s
    assert inst.conns["p1"].top == 1
    assert inst.conns["p1"].bot == 0
    assert inst.conns["p1"].width == 1
    assert isinstance(inst.conns["p8"], h.Concat)
    assert len(inst.conns["p8"].parts) == 4
    assert inst.conns["p8"].parts[0] is M2.s
    assert isinstance(inst.conns["p8"].parts[0], h.Signal)
    assert isinstance(inst.conns["p8"].parts[1], h.Slice)
    assert isinstance(inst.conns["p8"].parts[2], h.Slice)
    assert isinstance(inst.conns["p8"].parts[3], h.Slice)


def test_proto_roundtrip():
    # Test protobuf round-tripping

    # Create an empty (named) Module
    M1 = h.Module(name="M1")

    # Protobuf round-trip it
    ppkg = h.to_proto(M1)
    ns = h.from_proto(ppkg)

    assert isinstance(ns, SimpleNamespace)
    ns = ns.hdl21.tests.test_exports
    assert isinstance(ns, SimpleNamespace)
    assert isinstance(ns.M1, h.Module)
    assert len(ns.M1.signals) == 0
    assert len(ns.M1.ports) == 0
    assert len(ns.M1.instances) == 0


def test_proto_roundtrip2():
    # Create a child/leaf Module
    M1 = h.Module(name="M1")
    M1.i = h.Input()
    M1.o = h.Output()
    M1.p = h.Port()

    # Create an instantiating parent-module
    M2 = h.Module(name="M2")
    M2.i = h.Input()
    M2.o = h.Output()
    M2.p = h.Port()
    M2.s = h.Signal()

    # Add a few instances of it
    M2.i0 = M1()
    M2.i1 = M1()
    M2.i2 = M1()
    M2.i0(i=M2.i, o=M2.i1.i, p=M2.p)
    M2.i1(i=M2.i0.o, o=M2.s, p=M2.p)
    M2.i2(i=M2.s, o=M2.o, p=M2.p)

    # Protobuf round-trip it
    ppkg = h.to_proto(M2)
    ns = h.from_proto(ppkg)

    # And check all kinda stuff about what comes back
    assert isinstance(ns, SimpleNamespace)
    M1 = ns.hdl21.tests.test_exports.M1
    M2 = ns.hdl21.tests.test_exports.M2
    assert isinstance(M1, h.Module)
    assert isinstance(M2, h.Module)

    assert len(M1.signals) == 0
    assert len(M1.ports) == 3
    assert "i" in M1.ports
    assert M1.i.direction == h.signal.PortDir.INPUT
    assert M1.i.vis == h.signal.Visibility.PORT
    assert "o" in M1.ports
    assert M1.o.direction == h.signal.PortDir.OUTPUT
    assert M1.o.vis == h.signal.Visibility.PORT
    assert "p" in M1.ports
    assert M1.p.direction == h.signal.PortDir.NONE
    assert M1.p.vis == h.signal.Visibility.PORT
    assert len(M1.instances) == 0

    assert len(M2.instances) == 3
    for i in M2.instances.values():
        assert i.of is M1
        assert i.conns["p"] is M2.p
    assert len(M2.ports) == 3
    assert len(M2.signals) == 2
    assert "s" in M2.signals
    assert "i0_o" in M2.signals


def test_netlist_fmts():
    """Test netlisting basic types to several formats"""

    @h.module
    class Bot:
        s = h.Input(width=3)
        p = h.Output()

    @h.module
    class Top:
        s = h.Signal(width=3)
        p = h.Output()
        b = Bot(s=s, p=p)

    # Convert to proto
    ppkg = h.to_proto(Top)
    assert len(ppkg.modules) == 2
    assert ppkg.modules[0].name == "hdl21.tests.test_exports.Bot"
    assert ppkg.modules[1].name == "hdl21.tests.test_exports.Top"

    # Convert the proto-package to a netlist
    nl = StringIO()
    h.netlist(ppkg, nl, fmt="spectre")
    nl = nl.getvalue()

    # Basic checks on its contents
    assert "subckt Bot" in nl
    assert "s_2 s_1 s_0 p" in nl
    assert "subckt Top" in nl
    assert "b\n" in nl
    assert "( s_2 s_1 s_0 p )" in nl
    assert "+ p" in nl
    assert "+  Bot " in nl

    # Convert the proto-package to another netlist format
    nl = StringIO()
    h.netlist(ppkg, nl, fmt="verilog")
    nl = nl.getvalue()

    # Basic checks on its contents
    assert "module Bot" in nl
    assert "input wire [2:0] s," in nl
    assert "output wire p" in nl
    assert "endmodule // Bot" in nl
    assert "module Top" in nl
    assert ".s(s)" in nl
    assert ".p(p)" in nl
    assert "endmodule // Top" in nl

    nl = StringIO()
    h.netlist(ppkg, nl, fmt="spice")
    nl = nl.getvalue()
    assert ".SUBCKT Bot" in nl
    assert "+ s_2 s_1 s_0 p" in nl
    assert ".SUBCKT Top" in nl
    assert "+ p" in nl
    assert "xb" in nl
    assert "+ s_2 s_1 s_0 p" in nl
    assert "+ Bot" in nl


def test_spice_netlister():
    from hdl21.prefix import e

    @h.module
    class DUT:
        a = h.Input(width=5)
        b = h.Output(width=5)
        res = h.IdealResistor(h.ResistorParams(r=10 * e(3)))(p=a[0], n=b[0])
        cap = h.IdealCapacitor(h.IdealCapacitorParams(c=10 * e(-12)))(p=a[1], n=b[1])
        ind = h.IdealInductor(h.IdealInductorParams(l=10 * e(-9)))(p=a[2], n=b[2])

    ppkg = h.to_proto(DUT)
    nl = StringIO()
    h.netlist(ppkg, nl, fmt="spice")

    nl = nl.getvalue()
    assert ".SUBCKT DUT" in nl
    assert "+ a_4 a_3 a_2 a_1 a_0 b_4 b_3 b_2 b_1 b_0" in nl
    assert "rres" in nl
    assert "+ a_0 b_0" in nl
    assert "+ 10K" in nl
    assert "ccap" in nl
    assert "+ a_1 b_1" in nl
    assert "+ 10p" in nl
    assert "lind" in nl
    assert "+ a_2 b_2" in nl
    assert "+ 10n" in nl


def test_bad_proto_naming():
    """Test some naming cases which are alright in Python,
    but generate naming conflicts in serialization.
    Generally these boil down to "same `Module`-names in same Python module"."""

    # Create a naming conflict between Module definitions
    Ma1 = h.Module(name="Ma")
    Ma2 = h.Module(name="Ma")

    @h.module
    class MaParent:
        ma1 = Ma1()
        ma2 = Ma2()

    h.elaborate(MaParent)
    with pytest.raises(RuntimeError):
        h.to_proto(MaParent)

    # Do the same via some functions that should be generators
    def m1():
        return h.Module(name="Ma")

    def m2():
        return h.Module(name="Ma")

    @h.module
    class MaParent:
        ma1 = m1()()
        ma2 = m2()()

    h.elaborate(MaParent)
    with pytest.raises(RuntimeError):
        h.to_proto(MaParent)


def test_generator_recall():
    """Test multi-calling generators"""

    @h.generator
    def CallMeTwice(_: h.HasNoParams) -> h.Module:
        return h.Module()

    @h.module
    class Caller:
        m1 = CallMeTwice(h.NoParams)()
        m2 = CallMeTwice(h.NoParams)()

    # Convert to proto
    ppkg = h.to_proto(Caller)
    assert len(ppkg.modules) == 2
    assert ppkg.modules[0].name == "hdl21.tests.test_exports.CallMeTwice"
    assert ppkg.modules[1].name == "hdl21.tests.test_exports.Caller"

    # Convert the proto-package to a netlist, and run some (very basic) checks
    nl = StringIO()
    h.netlist(ppkg, nl)
    nl = nl.getvalue()
    assert "CallMeTwice" in nl
    assert "Caller" in nl

    # Round-trip back from Proto to Modules
    rt = h.from_proto(ppkg)
    ns = rt.hdl21.tests.test_exports
    assert isinstance(ns.Caller, h.Module)
    assert isinstance(getattr(ns, "CallMeTwice"), h.Module)


def test_roundtrip_external_module():
    """Test round-tripping `ExternalModule`s between Hdl21 and VLSIR Proto"""

    @h.paramclass
    class P:  # Our ExternalModule's parameter-type
        a = h.Param(dtype=int, desc="a", default=1)
        b = h.Param(dtype=str, desc="b", default="two")

    E = h.ExternalModule(name="E", port_list=[], paramtype=P)

    @h.module
    class HasE:
        e = E(P())()

    exported = h.to_proto(HasE)
    imported = h.from_proto(exported)
    h.to_proto(imported.hdl21.tests.test_exports.HasE)


def test_module_with_no_python_module():
    # Issue #48 https://github.com/dan-fritchman/Hdl21/issues/48
    exec("import hdl21 as h; h.to_proto(h.Module(name='not_in_a_pymodule'))")


def test_netlist_spicetypes():
    """# Test netlisting `ExternalModule`s with `SpiceType`s"""
    from hdl21.external_module import SpiceType
    from vlsirtools.netlist import NetlistFormat

    NmosModel = h.ExternalModule(
        name="nmos_model",
        port_list=deepcopy(h.Mos.port_list),
        paramtype=h.HasNoParams,
        spicetype=SpiceType.MOS,
    )

    @h.module
    class HasSpiceTypes:
        VSS = h.Signal()
        the_model_inst = NmosModel()(d=VSS, g=VSS, s=VSS, b=VSS)

    # Test netlisting in a handful of formats
    # Do some basic checking - mostly against exceptions.
    # But make sure the model-name got in there at least somewhere.

    sio = StringIO()
    h.netlist(HasSpiceTypes, sio, fmt=NetlistFormat.SPICE)
    nl = sio.getvalue()
    assert "nmos_model" in nl

    sio = StringIO()
    h.netlist(HasSpiceTypes, sio, fmt=NetlistFormat.XYCE)
    nl = sio.getvalue()
    assert "nmos_model" in nl

    sio = StringIO()
    h.netlist(HasSpiceTypes, sio, fmt=NetlistFormat.NGSPICE)
    nl = sio.getvalue()
    assert "nmos_model" in nl

    sio = StringIO()
    h.netlist(HasSpiceTypes, sio, fmt=NetlistFormat.SPECTRE)
    nl = sio.getvalue()
    assert "nmos_model" in nl


def test_module_with_literals():
    """# Test exporting modules with literals"""

    @h.module
    class HasLit:
        a, b, c = h.Ports(3)

    # Add the literals
    HasLit.literals.extend(
        [
            h.Literal("generate some_terrible_verilog_code"),
            h.Literal(".some_spice_attribute what=ever"),
            h.Literal("PRAGMA: some_pragma"),
        ]
    )

    # Test converting to proto
    pkg = h.to_proto(HasLit)
    pmod = pkg.modules[0]
    assert isinstance(pmod, vlsir.circuit.Module)
    assert pmod.name == "hdl21.tests.test_exports.HasLit"
    assert len(pmod.ports) == 3
    assert len(pmod.signals) == 3

    assert pmod.literals == [l.text for l in HasLit.literals]
    assert pmod.literals == [
        "generate some_terrible_verilog_code",
        ".some_spice_attribute what=ever",
        "PRAGMA: some_pragma",
    ]

    dest = StringIO()
    h.netlist(HasLit, dest)
    assert "generate some_terrible_verilog_code" in dest.getvalue()
    assert ".some_spice_attribute what=ever" in dest.getvalue()
    assert "PRAGMA: some_pragma" in dest.getvalue()

    # Test round-tripping
    ns = h.from_proto(pkg)
    assert isinstance(ns, SimpleNamespace)
    ns = ns.hdl21.tests.test_exports
    assert isinstance(ns, SimpleNamespace)
    HasLitRoundTripped = ns.HasLit
    assert isinstance(HasLitRoundTripped, h.Module)
    assert len(HasLitRoundTripped.literals) == 3
    assert HasLitRoundTripped.literals == HasLit.literals


def test_external_module_to_vlsir():
    emod = h.ExternalModule(
        name="emod",
        port_list=[h.Port(name="p", direction=h.PortDir.INPUT)],
    )
    pmod = h.proto.export_external_module(emod)
    assert isinstance(pmod, vlsir.circuit.ExternalModule)
    # FIXME: some better tests



"""
# Test Content

A semi-comprehensive set of Hdl21 features, for use in testing related packages, e.g. PDKs. 
"""

import hdl21 as h


@h.bundle
class Abc:
    a, b, c = h.Ports(3)


@h.paramclass
class N:
    n = h.Param(dtype=int, desc="n")


@h.module
class Bot:
    abc = Abc(port=True)
    a, b, c = h.Ports(3, width=2)


@h.generator
def Inner(params: N) -> h.Module:
    Empty = h.Module(name="Empty")
    EmptyEmod = h.ExternalModule(name="EmptyEmod", port_list=[])

    @h.module
    class Inner:
        abc = Abc()
        a, b, c = h.Signals(3, width=2)
        bot = Bot(abc=abc, a=a, b=b, c=h.Concat(c[1], c[0]))
        emod_call = EmptyEmod()()
        empty_array = params.n * Empty()

    return Inner


@h.generator
def Top1(_: N) -> h.Module:
    m = h.Module()
    m.i = Inner(n=5)()
    return m


@h.module
class Top2:
    i = Inner(n=5)


def walker_test_content():
    """Create a set of content for a `h.HierarchyWalker` test.
    Returns a list of combinations of modules, generator calls, and all the internal features inside."""
    return [Top1(N(1)), Top2]



import pytest

import hdl21 as h
from hdl21.flatten import flatten, is_flat, walk


@h.module
class Inverter:
    vdd, vss, vin, vout = h.Ports(4)
    pmos = h.Pmos()(d=vout, g=vin, s=vdd, b=vdd)
    nmos = h.Nmos()(d=vout, g=vin, s=vss, b=vss)


@h.module
class Buffer:
    vdd, vss, vin, vout = h.Ports(4)
    inv_1 = Inverter(vdd=vdd, vss=vss, vin=vin)  # type: ignore
    inv_2 = Inverter(vdd=vdd, vss=vss, vin=inv_1.vout, vout=vout)  # type: ignore


@h.module
class DoubleBuffer:
    vdd, vss, vin, vout = h.Ports(4)
    buffer_1 = Buffer(vdd=vdd, vss=vss, vin=vin)
    buffer_2 = Buffer(vdd=vdd, vss=vss, vin=buffer_1.vout, vout=vout)


@h.module
class InvBuffer:
    vdd, vss, vin, vout = h.Ports(4)
    inv = Inverter(vdd=vdd, vss=vss, vin=vin)
    buffer = Buffer(vdd=vdd, vss=vss, vin=inv.vout, vout=vout)


def test_is_flat():
    assert is_flat(Inverter)
    assert not is_flat(Buffer)
    assert not is_flat(InvBuffer)


def test_flatten_inv():
    inv = Inverter
    assert is_flat(inv)
    inv_raw_proto = h.to_proto(h.elaborate(inv))
    inv_flatten_proto = h.to_proto(flatten(inv))
    assert inv_raw_proto == inv_flatten_proto


def test_flatten_buffer():
    buffer = Buffer
    assert not is_flat(buffer)
    buffer_flat = flatten(buffer)
    assert buffer_flat.instances.keys() == {
        "inv_1:pmos",
        "inv_1:nmos",
        "inv_2:pmos",
        "inv_2:nmos",
    }
    assert buffer_flat.ports.keys() == {"vdd", "vss", "vin", "vout"}
    assert buffer_flat.signals.keys() == {"inv_1_vout"}
    assert is_flat(buffer_flat)


def test_flatten_double_buffer():
    double_buffer = DoubleBuffer
    assert not is_flat(double_buffer)
    double_buffer_flat = flatten(double_buffer)
    assert double_buffer_flat.instances.keys() == {
        "buffer_1:inv_1:pmos",
        "buffer_1:inv_1:nmos",
        "buffer_1:inv_2:pmos",
        "buffer_1:inv_2:nmos",
        "buffer_2:inv_1:pmos",
        "buffer_2:inv_1:nmos",
        "buffer_2:inv_2:pmos",
        "buffer_2:inv_2:nmos",
    }
    assert double_buffer_flat.ports.keys() == {"vdd", "vss", "vin", "vout"}
    assert double_buffer_flat.signals.keys() == {
        "buffer_1_vout",  # connection between two buffers
        "buffer_1:inv_1_vout",  # internal signal between two inverters in buffer 1
        "buffer_2:inv_1_vout",  # internal signal between two inverters in buffer 2
    }
    assert is_flat(double_buffer_flat)


def test_flatten_inv_buffer():
    inv_buffer = InvBuffer
    assert not is_flat(inv_buffer)
    inv_buffer_flat = flatten(inv_buffer)
    assert inv_buffer_flat.instances.keys() == {
        "inv:pmos",
        "inv:nmos",
        "buffer:inv_1:pmos",
        "buffer:inv_1:nmos",
        "buffer:inv_2:pmos",
        "buffer:inv_2:nmos",
    }
    assert inv_buffer_flat.ports.keys() == {"vdd", "vss", "vin", "vout"}
    assert inv_buffer_flat.signals.keys() == {"inv_vout", "buffer:inv_1_vout"}
    assert is_flat(inv_buffer_flat)


def test_flatten_node_desc():
    m = h.elaborate(Buffer)
    nodes = list(walk(m, parents=[]))
    assert (
        str(nodes[0])
        == "{'name': 'inv_1:pmos', 'path': ['inv_1', 'pmos'], 'conns': {'d': 'inv_1_vout', 'g': 'vin', 's': 'vdd', 'b': 'vdd'}}"
    )


@pytest.mark.xfail(reason="FIXME: flatten with slices & concats")
def test_flatten_with_slices():
    """Flatten a Module with slices"""

    @h.module
    class InvTwoPack:
        # A two-bit-bus's worth of inverters
        vdd, vss = h.Ports(2)
        vin, vout = 2 * h.Port(width=2)
        inv_0 = Inverter(vdd=vdd, vss=vss, vin=vin[0], vout=vout[0])  # type: ignore
        inv_1 = Inverter(vdd=vdd, vss=vss, vin=vin[1], vout=vout[1])  # type: ignore

    flattened = flatten(InvTwoPack)


@pytest.mark.xfail(reason="FIXME: flatten with slices & concats")
def test_flatten_with_concat():
    """Test flattening a module with signal concatenations."""

    @h.module
    class NmosArray:
        # An instance array of Nmos'es. Maybe for a current DAC?
        g, vss = 2 * h.Port()
        d = h.Port(width=8)
        nmoses = 8 * h.Nmos()(g=g, d=d, s=vss, b=vss)

    @h.module
    class M:
        # A thing that instantiates an NmosArray, and concatenates some signals together to form its drain connections.
        g, vss = 2 * h.Port()
        s4 = h.Signal(width=4)
        s2 = h.Signal(width=2)
        s1 = h.Signal(width=1)
        s0 = h.Signal(width=1)
        nmos_array = NmosArray(d=h.Concat(s4, s2, s1, s0), g=g, vss=vss)

    flattened = flatten(M)



"""
# Hdl21 Sample PDK 

Unit Tests
"""

import hdl21 as h
from io import StringIO
from . import pdk as sample_pdk


def test_default():
    h.pdk.set_default(sample_pdk)
    assert h.pdk.default() is sample_pdk


def mosmodule():
    # Create a simple Module with each of the default-param Mos types
    hasmos = h.Module(name="hasmos")
    z = hasmos.z = h.Signal()
    hasmos.n = h.Nmos(h.MosParams())(d=z, g=z, s=z, b=z)
    hasmos.p = h.Pmos(h.MosParams())(d=z, g=z, s=z, b=z)

    # Checks on the initial Module
    assert isinstance(hasmos.n, h.Instance)
    assert isinstance(hasmos.p, h.Instance)
    assert isinstance(hasmos.n.of, h.PrimitiveCall)
    assert isinstance(hasmos.p.of, h.PrimitiveCall)

    return hasmos


def test_compile():
    hasmos = mosmodule()

    # Compile it for the PDK
    sample_pdk.compile(hasmos)

    # And check what came back
    assert isinstance(hasmos.n, h.Instance)
    assert isinstance(hasmos.p, h.Instance)
    assert isinstance(hasmos.n.of, h.ExternalModuleCall)
    assert isinstance(hasmos.p.of, h.ExternalModuleCall)
    assert isinstance(hasmos.n.of.params, sample_pdk.SamplePdkMosParams)
    assert isinstance(hasmos.p.of.params, sample_pdk.SamplePdkMosParams)


def test_netlist():
    hasmos = mosmodule()

    # Netlist it for the PDK
    sample_pdk.compile(hasmos)
    h.netlist(hasmos, StringIO(), fmt="spice")
    h.netlist(hasmos, StringIO(), fmt="spectre")
    h.netlist(hasmos, StringIO(), fmt="verilog")



""" 
# `InstanceBundle`s ElabPass Pass

Transform each `InstanceBundle` into a set of "scalar" `Instance`s, 
and reconnect them to the signals of the original `Bundle`. 

This pass is designed to execute relatively early, and relies on `BundleRef`s in particular to 
be resolved *after* it completes. 

FIXME: sort out the relationship between this pass and `ConnTypes`. 
       Currently this pass checks for "bundle type identity" (i.e. `conn.bundle IS inst.bundle`), 
       and `ConnTypes` checks for "structural equivalence". 
"""


# Local imports
from ...connect import is_connectable
from ...bundle import AnonymousBundle, BundleInstance, _bundle_ref
from ...module import Module
from ...instance import Instance, InstanceBundle

# Import the base class
from .base import ElabPass


class InstBundleElabPass(ElabPass):
    """
    # Instance Bundle Elaboration Pass
    """

    def elaborate_module(self, module: Module) -> Module:
        """Elaborate a Module
        Remove each Instance Bundle, and replace it with "scalar" Instances."""

        while module.instbundles:
            name, inst = module.instbundles.popitem()
            module.namespace.pop(name)
            self.elaborate_instance_bundle(module, inst)

        return module

    def elaborate_instance_bundle(
        self, module: Module, instbundle: InstanceBundle
    ) -> InstanceBundle:
        """# Elaborate an Instance Bundle
        Replace it with "scalar" Instances and reconnect them."""

        if len(instbundle.bundle.bundles):
            msg = f"Invalid Instance Bundle {instbundle} with nested Bundles"
            self.fail(msg)

        # Create a replacement instance for each Signal in the Bundle
        # Keep a mapping from signal-names to those instances for forthcoming connections
        signal_names_to_instances = {
            signame: module.add(
                name=self.flatname(
                    segments=[instbundle.name, signame], avoid=module.namespace
                ),
                val=Instance(of=instbundle.of),
            )
            for signame in instbundle.bundle.signals
        }

        # Disconnect each port from the `InstanceBundle`, and map it into the new scalar `Instance`s
        for portname in list(instbundle.conns.keys()):
            conn = instbundle.disconnect(portname)

            if isinstance(conn, BundleInstance):
                # If the connection is a Bundle instance, connect each new instance to the paired-named Signals
                if conn.of is not instbundle.bundle:
                    msg = f"Invalid Instance Bundle connection between {conn.of} and {instbundle.bundle}"
                    self.fail(msg)

                for signame, new_inst in signal_names_to_instances.items():
                    new_inst.connect(portname, _bundle_ref(conn, signame))

            elif isinstance(conn, AnonymousBundle):
                for signame, new_inst in signal_names_to_instances.items():
                    new_inst.connect(portname, conn.get(signame))

            elif is_connectable(conn):
                # If the connection is a scalar, connect it to each new instance
                for new_inst in signal_names_to_instances.values():
                    new_inst.connect(portname, conn)

            else:
                self.fail(
                    f"Invalid connection {conn} for `InstanceBundle` {instbundle.name}"
                )

        return instbundle



"""
# hdl21 Signals and Ports 

Signals are Hdl21's base-level unit of hardware connectivity. 

Each `Signal` is analogous in content to a *bus* or (single-dimensional) *array* in many legacy HDLs. 
(Most similar to Verilog's packed single-dimensional arrays.) 
The `Signal.width` field indicates the bit-width of said bus. 
It defaults to one for scalar Signals. 
Widths of zero or less generate errors, both at construction-time and later. 

Hdl21 `Signals` are *untyped*. 
They represent a connection, not the type or value of data it carries. 
In this sense they are more similar to analog-style environments than to most legacy HDLs. 

`Ports`, `Inputs`, `Outputs`, and `Inouts` are not dedicated Hdl21 types, 
but thin convenience function-wrappers around `Signal`. 
Each `Signal` includes enumerated fields for its visibility (internal vs port) 
and direction. For internal `Signals`, the `direction` field is globally expected to be ignored. 

"""

from copy import copy
from enum import Enum
from dataclasses import field
from typing import Callable, Optional, List, Set, Tuple

# Local imports
from .role import Role
from .datatype import datatype
from .visibility import Visibility
from .connect import connectable
from .sliceable import sliceable
from .concat import concatable
from .props import Properties


class PortDir(Enum):
    """# Port Direction Enumeration"""

    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    INOUT = "INOUT"
    NONE = "NONE"  # No direction or unspecified

    def flipped(self) -> "PortDir":
        """# Flip the direction of a port"""
        if self == PortDir.INPUT:
            return PortDir.OUTPUT
        if self == PortDir.OUTPUT:
            return PortDir.INPUT
        return self  # INOUT or NONE


class Usage(Enum):
    """# Signal Usage"""

    SIGNAL = "SIGNAL"
    POWER = "POWER"
    GROUND = "GROUND"
    CLOCK = "CLOCK"


@sliceable
@concatable
@connectable
@datatype
class Signal:
    """
    # Signal

    Signals are Hdl21's base-level unit of hardware connectivity.

    Each `Signal` is analogous in content to a *bus* or (single-dimensional) *array* in many legacy HDLs.
    (Most similar to Verilog's packed single-dimensional arrays.)
    The `Signal.width` field indicates the bit-width of said bus.
    It defaults to one for scalar Signals.
    Widths of zero or less generate errors, both at construction-time and later.

    Hdl21 `Signals` are *untyped*.
    They represent a connection, not the type or value of data it carries.
    In this sense they are more similar to analog-style environments than to most legacy HDLs.

    `Ports`, `Inputs`, `Outputs`, and `Inouts` are not dedicated Hdl21 types,
    but thin convenience function-wrappers around `Signal`.
    Each `Signal` includes enumerated fields for its visibility (internal vs port)
    and direction. For internal `Signals`, the `direction` field is globally expected to be ignored.

    """

    name: Optional[str] = None  # Signal name
    width: int = 1  # Bit-width
    vis: Visibility = field(repr=False, default=Visibility.INTERNAL)  # Port visibility
    direction: PortDir = field(repr=False, default=PortDir.NONE)  # Port direction
    usage: Usage = field(repr=False, default=Usage.SIGNAL)  # Signal usage
    props: Properties = field(repr=False, default_factory=Properties)  # Properties
    desc: Optional[str] = None  # Description
    src: Optional[Role] = field(repr=False, default=None)  # Source Role
    dest: Optional[Role] = field(repr=False, default=None)  # Destination Role

    # Related signals
    related_clk: Optional["Signal"] = field(repr=False, default=None)
    # Related clock signal
    related_pwr: Optional["Signal"] = field(repr=False, default=None)
    # Related power signal
    related_gnd: Optional["Signal"] = field(repr=False, default=None)
    # Related ground signal

    def __post_init__(self):
        if self.width < 1:
            raise ValueError(f"Signal {self.name} width must be positive")
        self._parent_module: Optional["Module"] = None
        self._slices: Set["Slice"] = set()
        self._concats: Set["Concat"] = set()
        self._connected_ports: Set["PortRef"] = set()

        # Back-references to related signals
        self._related_clk_of: Set["Signal"] = set()
        self._related_pwr_of: Set["Signal"] = set()
        self._related_gnd_of: Set["Signal"] = set()

        # Add those back-references to signals *we* relate to.
        # Note this only happens at construction-time.
        if self.related_clk is not None:
            self.related_clk._related_clk_of.add(self)
        if self.related_pwr is not None:
            self.related_pwr._related_pwr_of.add(self)
        if self.related_gnd is not None:
            self.related_gnd._related_gnd_of.add(self)

    def __eq__(self, other) -> bool:
        # Identity is equality
        return other is self

    def __hash__(self) -> bool:
        # Identity is equality
        return hash(id(self))

    def __copy__(self) -> "Signal":
        """Signal copying implementation
        Keeps "public" fields such as name and width,
        while dropping "per-module" fields such as `_slices`."""
        # Notably `_parent_module` *is not* copied.
        # It will generally be set when the copy is added to any new Module.
        return Signal(
            name=self.name,
            width=self.width,
            vis=self.vis,
            direction=self.direction,
            desc=self.desc,
            src=self.src,
            dest=self.dest,
        )

    def __deepcopy__(self, _memo) -> "Signal":
        """Signal "deep" copies"""
        # The same as shallow ones; there is no "deep" data being copied.
        return self.__copy__()

    def __rmul__(self, num: int) -> List["Signal"]:
        """# Right multiplication. Creates `num` copies of this Signal."""
        if not isinstance(num, int):
            return NotImplemented
        return [copy(self) for _ in range(num)]


"""
# Constructor Helpers 

Thin wrappers around the `Signal` constructor which set common fields, 
e.g. `Port()` replacing `Signal(vis=Visibility.PORT)`.
"""


def Input(**kwargs) -> Signal:
    """# Input Port Constructor
    Thin wrapper around `hdl21.Signal`"""
    return Signal(vis=Visibility.PORT, direction=PortDir.INPUT, **kwargs)


def Output(**kwargs) -> Signal:
    """# Output Port Constructor
    Thin wrapper around `hdl21.Signal`"""
    return Signal(vis=Visibility.PORT, direction=PortDir.OUTPUT, **kwargs)


def Inout(**kwargs) -> Signal:
    """# Inout Port Constructor
    Thin wrapper around `hdl21.Signal`"""
    return Signal(vis=Visibility.PORT, direction=PortDir.INOUT, **kwargs)


def Port(direction=PortDir.NONE, **kwargs) -> Signal:
    """# Port Constructor
    Thin wrapper around `hdl21.Signal`.
    The `direction` argument sets the Port's direction,
    and defaults to the unknown direction `PortDir.NONE`."""
    return Signal(direction=direction, vis=Visibility.PORT, **kwargs)


def Power(direction=PortDir.INPUT, **kwargs) -> Signal:
    """# Power Signal Constructor
    Thin wrapper around `hdl21.Signal`, setting the `POWER` usage attribute.
    `Power` by default returns a `Port` with `PortDir.INPUT` direction.
    Its direction can be overrident via the `direction` argument."""
    return Signal(direction=direction, vis=Visibility.PORT, usage=Usage.POWER, **kwargs)


def Ground(direction=PortDir.INPUT, **kwargs) -> Signal:
    """# Ground Signal Constructor
    Thin wrapper around `hdl21.Signal`, setting the `GROUND` usage attribute.
    `Ground` by default returns a `Port` with `PortDir.INPUT` direction.
    Its direction can be overrident via the `direction` argument."""
    return Signal(
        direction=direction, vis=Visibility.PORT, usage=Usage.GROUND, **kwargs
    )


def PowerGround(direction=PortDir.INPUT, **kwargs) -> Tuple[Signal, Signal]:
    """# Power and Ground Signal Constructor
    Returns a tuple of the two in (Power, Ground) order."""
    return (Power(direction=direction, **kwargs), Ground(direction=direction, **kwargs))


def Clock(direction=PortDir.INPUT, **kwargs) -> Signal:
    """# Clock Signal Constructor
    Thin wrapper around `hdl21.Signal`, setting the `CLOCK` usage attribute.
    `Clock` by default returns a `Port` with `PortDir.INPUT` direction.
    Its direction can be overrident via the `direction` argument."""
    return Signal(direction=direction, vis=Visibility.PORT, usage=Usage.CLOCK, **kwargs)


"""
# Plural Constructors

Wrappers that generate the functions named `Signals`, `Ports`, and the like. 
"""


def _pluralize(fn: Callable):
    """Inner helper method for creating "plural" versions of `Signal` constructors."""

    def _plural(num: int, **kwargs) -> List[Signal]:
        return [fn(**kwargs) for _ in range(num)]

    # Give the wrapper a plural name, e.g. "Signals" or "Clocks"
    _plural.__name__ = fn.__name__ + "s"
    # And a basic doc string
    _plural.__doc__ = f"Create `num` new {fn.__name__}s."
    # Add the wrapper to this module's namespace
    globals()[_plural.__name__] = _plural


# Create all those plural versions
[_pluralize(fn) for fn in [Signal, Port, Input, Output, Inout, Power, Ground, Clock]]


"""
# Non-public (well, at least in intent) Methods
"""


def _copy_to_internal(sig: Signal) -> Signal:
    """Make a copy of `sig`, replacing its visibility and port-direction to be internal."""
    sig = copy(sig)
    sig.vis = Visibility.INTERNAL
    sig.direction = PortDir.NONE
    sig._parent_module = None
    return sig



"""
# Hdl21 Primitive Modules

Primitives are leaf-level Modules typically defined not by users, 
but by simulation tools or device fabricators. 
Prominent examples include MOS transistors, diodes, resistors, and capacitors. 

Primitives divide in two classes, `physical` and `ideal`, indicated by their `primtype` attribute. 
`PrimitiveType.IDEAL` primitives specify circuit-theoretic ideal elements 
e.g. resistors, capacitors, inductors, and notably aphysical elements 
such as ideal voltage and current sources. 

`PrimitiveType.PHYSICAL` primitives in contrast specify abstract versions 
of ultimately physically-realizable elements such as transistors and diodes. 
These elements typically require some external translation, e.g. by a process-technology 
library, to execute in simulations or to be realized in hardware. 

Many element-types (particularly passives) come in both `ideal` and `physical` flavors, 
as typical process-technologies include physical passives, but with far different 
parameterization than ideal passives. For example resistors are commonly specified 
in physical length and width. Capacitors are similarly specified in physical terms, 
often adding metal layers or other physical features. The component-value (R,C,L, etc.) 
for these physically-specified cells is commonly suggestive or optional. 

Both the `Primitive` type and all of its instances are defined in this module. 
The collection of `Primitive` instances is often referred to as Hdl21's "primitive library". 

Summary of the content of the primitive library: 

| Name                           | Description                       | Type     | Aliases                               | Ports        |
| ------------------------------ | --------------------------------- | -------- | ------------------------------------- | ------------ |
| Mos                            | Mos Transistor                    | PHYSICAL | MOS                                   | d, g, s, b   |
| IdealResistor                  | Ideal Resistor                    | IDEAL    | R, Res, Resistor, IdealR, IdealRes    | p, n         |
| PhysicalResistor               | Physical Resistor                 | PHYSICAL | PhyR, PhyRes, ResPhy, PhyResistor     | p, n         |
| ThreeTerminalResistor          | Three Terminal Resistor           | PHYSICAL | Res3, PhyRes3, ResPhy3, PhyResistor3  | p, n, b      |
| IdealCapacitor                 | Ideal Capacitor                   | IDEAL    | C, Cap, Capacitor, IdealC, IdealCap   | p, n         |
| PhysicalCapacitor              | Physical Capacitor                | PHYSICAL | PhyC, PhyCap, CapPhy, PhyCapacitor    | p, n         |
| ThreeTerminalCapacitor         | Three Terminal Capacitor          | PHYSICAL | Cap3, PhyCap3, CapPhy3, PhyCapacitor3 | p, n, b      |
| IdealInductor                  | Ideal Inductor                    | IDEAL    | L, Ind, Inductor, IdealL, IdealInd    | p, n         |
| PhysicalInductor               | Physical Inductor                 | PHYSICAL | PhyL, PhyInd, IndPhy, PhyInductor     | p, n         |
| ThreeTerminalInductor          | Three Terminal Inductor           | PHYSICAL | Ind3, PhyInd3, IndPhy3, PhyInductor3  | p, n, b      |
| PhysicalShort                  | Short-Circuit/ Net-Tie            | PHYSICAL | Short                                 | p, n         |
| DcVoltageSource                | DC Voltage Source                 | IDEAL    | V, Vdc, Vsrc                          | p, n         |
| PulseVoltageSource             | Pulse Voltage Source              | IDEAL    | Vpu, Vpulse                           | p, n         |
| CurrentSource                  | Ideal DC Current Source           | IDEAL    | I, Idc, Isrc                          | p, n         |
| VoltageControlledVoltageSource | Voltage Controlled Voltage Source | IDEAL    | Vcvs, VCVS                            | p, n, cp, cn |
| CurrentControlledVoltageSource | Current Controlled Voltage Source | IDEAL    | Ccvs, CCVS                            | p, n, cp, cn |
| VoltageControlledCurrentSource | Voltage Controlled Current Source | IDEAL    | Vccs, VCCS                            | p, n, cp, cn |
| CurrentControlledCurrentSource | Current Controlled Current Source | IDEAL    | Cccs, CCCS                            | p, n, cp, cn |
| Bipolar                        | Bipolar Transistor                | PHYSICAL | Bjt, BJT                              | c, b, e      |
| Diode                          | Diode                             | PHYSICAL | D                                     | p, n         |

"""

# Non-docstring comment:
#
# That big table above is also included in the Hdl21 package documentation.
# It is generated by the script `Hdl21/scripts/primtable.py`,
# which imports the contents of this module and prints a line per primitive.
# On changes to this module, re-run the script, paste the table here and anywhere else it is used.

# Std-Lib Imports
import copy
from enum import Enum
from dataclasses import replace
from typing import Optional, Any, List, Type, Dict

# PyPi Imports
from pydantic.dataclasses import dataclass

# Local imports
from .default import Default
from .call import param_call
from .params import paramclass, Param, isparamclass, NoParams, _unique_name
from .signal import Port, Signal, Visibility
from .instance import calls_instantiate
from .scalar import Scalar


class PrimitiveType(Enum):
    """Enumerated Primitive-Types"""

    IDEAL = "IDEAL"
    PHYSICAL = "PHYSICAL"


@dataclass
class Primitive:
    """# Hdl21 Primitive Component

    Primitives are leaf-level Modules typically defined not by users,
    but by simulation tools or device fabricators.
    Prominent examples include MOS transistors, diodes, resistors, and capacitors.
    """

    name: str  # Primitive Name
    desc: str  # String Description
    port_list: List[Signal]  # Ordered Port List
    paramtype: Type[object]  # Class/ Type of valid Parameters
    primtype: PrimitiveType  # Ideal vs Physical Primitive-Type

    def __post_init__(self):
        """After type-checking, do plenty more checks on values"""
        if not isparamclass(self.paramtype):
            msg = f"Invalid Primitive param-type {self.paramtype} for {self.name}, must be an `hdl21.paramclass`"
            raise TypeError(msg)
        for p in self.port_list:
            if not p.name:
                raise ValueError(f"Unnamed Primitive Port {p} for {self.name}")
            if p.vis != Visibility.PORT:
                msg = f"Invalid Primitive Port {p.name} on {self.name}; must have PORT visibility"
                raise ValueError(msg)

    def __call__(self, arg: Any = Default, **kwargs) -> "PrimitiveCall":
        params = param_call(callee=self, arg=arg, **kwargs)
        return PrimitiveCall(prim=self, params=params)

    @property
    def Params(self) -> Type:
        """Type-style alias for the parameter-type."""
        return self.paramtype

    @property
    def ports(self) -> Dict[str, Signal]:
        return {p.name: p for p in self.port_list}

    def __eq__(self, other) -> bool:
        # Identity is equality
        return other is self

    def __hash__(self) -> bool:
        # Identity is equality
        return hash(id(self))


@calls_instantiate
@dataclass
class PrimitiveCall:
    """Primitive Call
    A combination of a Primitive and its Parameter-values,
    typically generated by calling the Primitive."""

    prim: Primitive
    params: Any = NoParams

    def __post_init__(self):
        # Type-validate our parameters
        if not isinstance(self.params, self.prim.paramtype):
            msg = f"Invalid parameters {self.params} for Primitive {self.prim}. Must be {self.prim.paramtype}"
            raise TypeError(msg)

    @property
    def name(self) -> str:
        return self.prim.name + "(" + _unique_name(self.params) + ")"

    @property
    def ports(self) -> dict:
        return self.prim.ports

    def __eq__(self, other) -> bool:
        """Call equality requires:
        * *Identity* between prims, and
        * *Equality* between parameter-values."""
        return self.prim is other.prim and self.params == other.params

    def __hash__(self):
        """Generator-Call hashing, consistent with `__eq__` above, uses:
        * *Identity* of its prim, and
        * *Value* of its parameters.
        The two are joined for hashing as a two-element tuple."""
        return hash((id(self.prim), self.params))


@dataclass
class PrimLibEntry:
    """# Entry in the Primitive Library"""

    prim: Primitive
    aliases: List[str]


# Dictionary storing all primitives, keyed by their primary name.
# Stores aliases on the side.
_primitives: Dict[str, PrimLibEntry] = dict()


def _add(prim: Primitive, aliases: List[str]) -> Primitive:
    """Add a primitive to this library.
    Ensures its identifier matches its `name` field, and adds any aliases to the global namespace.
    This is a private function and should be used solely during `hdl21.primitives` import-time.
    """
    global _primitives

    if prim.name in _primitives or prim.name in globals():
        raise ValueError(f"Duplicate primitive name {prim.name}")

    # Add the combination as a new entry
    entry = PrimLibEntry(prim, aliases)
    _primitives[prim.name] = entry
    globals()[prim.name] = prim

    for alias in aliases:
        if alias in _primitives or alias in globals():
            raise ValueError(f"Duplicate primitive alias {alias}")
        globals()[alias] = prim
    return prim


""" 
Mos Transistor Section 
"""


class MosType(Enum):
    """# MOS Type (NMOS/ PMOS) Enumeration"""

    NMOS = "NMOS"
    PMOS = "PMOS"


class MosVth(Enum):
    """# MOS Threshold Enumeration"""

    STD = "STD"
    LOW = "LOW"
    HIGH = "HIGH"
    ULTRA_LOW = "ULTRA_LOW"
    ULTRA_HIGH = "ULTRA_HIGH"
    ZERO = "ZERO"
    NATIVE = "NATIVE"


class MosFamily(Enum):
    """# MOS Family Enumeration"""

    NONE = "NONE"
    CORE = "CORE"
    IO = "IO"
    LP = "LP"
    HP = "HP"
    RF = "RF"


@paramclass
class MosParams:
    """# MOS Transistor Parameters"""

    # All optional
    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
    nf = Param(dtype=Optional[Scalar], desc="Number of parallel fingers", default=None)
    mult = Param(dtype=Optional[Scalar], desc="Multiplier", default=None)

    tp = Param(dtype=MosType, desc="MosType (Nmos/ Pmos)", default=MosType.NMOS)
    vth = Param(dtype=MosVth, desc="Threshold voltage specifier", default=MosVth.STD)
    family = Param(dtype=MosFamily, desc="Device family", default=MosFamily.NONE)
    model = Param(dtype=Optional[str], desc="Model (Name)", default=None)

    # def __post_init__(self):
    #     """Value Checks"""
    #     # FIXME: re-introduce these, for the case in which the parameters are `Prefixed` and not `Literal` values.
    #     if self.w <= 0:
    #         raise ValueError(f"MosParams with invalid width {self.w}")
    #     if self.l <= 0:
    #         raise ValueError(f"MosParams with invalid length {self.l}")
    #     if self.nf <= 0:
    #         msg = f"MosParams with invalid number parallel fingers {self.nf}"
    #         raise ValueError(msg)


# Mos Transistor Ports, in SPICE Conventional Order
MosPorts = [
    Port(name="d", desc="Drain"),
    Port(name="g", desc="Gate"),
    Port(name="s", desc="Source"),
    Port(name="b", desc="Bulk"),
]

Mos = _add(
    prim=Primitive(
        name="Mos",
        desc="Mos Transistor",
        port_list=copy.deepcopy(MosPorts),
        paramtype=MosParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["MOS"],
)


def Nmos(arg: Any = Default, **kwargs) -> Primitive:
    """Nmos Constructor. A thin wrapper around `hdl21.primitives.Mos`"""
    mos = Mos(arg, **kwargs)
    mos.params = replace(mos.params, tp=MosType.NMOS)
    return mos


def Pmos(arg: Any = Default, **kwargs) -> Primitive:
    """Pmos Constructor. A thin wrapper around `hdl21.primitives.Mos`"""
    mos = Mos(arg, **kwargs)
    mos.params = replace(mos.params, tp=MosType.PMOS)
    return mos


""" 
Passives
"""

# Oft-reused port list for the passive elements
PassivePorts = [Port(name="p"), Port(name="n")]
# And the three-terminal version
ThreeTerminalPorts = [Port(name="p"), Port(name="n"), Port(name="b")]


@paramclass
class ResistorParams:
    r = Param(dtype=Scalar, desc="Resistance (ohms)")


IdealResistor = _add(
    prim=Primitive(
        name="IdealResistor",
        desc="Ideal Resistor",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=ResistorParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["R", "Res", "Resistor", "IdealR", "IdealRes"],
)


@paramclass
class PhysicalResistorParams:
    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
    model = Param(dtype=Optional[str], desc="Model (Name)", default=None)


PhysicalResistor = _add(
    prim=Primitive(
        name="PhysicalResistor",
        desc="Physical Resistor",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=PhysicalResistorParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["PhyR", "PhyRes", "ResPhy", "PhyResistor"],
)


ThreeTerminalResistor = _add(
    prim=Primitive(
        name="ThreeTerminalResistor",
        desc="Three Terminal Resistor",
        port_list=copy.deepcopy(ThreeTerminalPorts),
        paramtype=PhysicalResistorParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["Res3", "PhyRes3", "ResPhy3", "PhyResistor3"],
)


@paramclass
class IdealCapacitorParams:
    c = Param(dtype=Scalar, desc="Capacitance (F)")


IdealCapacitor = _add(
    prim=Primitive(
        name="IdealCapacitor",
        desc="Ideal Capacitor",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=IdealCapacitorParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["C", "Cap", "Capacitor", "IdealC", "IdealCap"],
)


@paramclass
class PhysicalCapacitorParams:
    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
    c = Param(dtype=Optional[Scalar], desc="Capacitance (F)", default=None)
    model = Param(dtype=Optional[str], desc="Model (Name)", default=None)
    mult = Param(dtype=Optional[str], desc="Multiplier", default=None)


PhysicalCapacitor = _add(
    prim=Primitive(
        name="PhysicalCapacitor",
        desc="Physical Capacitor",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=PhysicalCapacitorParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["PhyC", "PhyCap", "CapPhy", "PhyCapacitor"],
)


ThreeTerminalCapacitor = _add(
    prim=Primitive(
        name="ThreeTerminalCapacitor",
        desc="Three Terminal Capacitor",
        port_list=copy.deepcopy(ThreeTerminalPorts),
        paramtype=PhysicalCapacitorParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["Cap3", "PhyCap3", "CapPhy3", "PhyCapacitor3"],
)


@paramclass
class IdealInductorParams:
    l = Param(dtype=Scalar, desc="Inductance (H)")


IdealInductor = _add(
    prim=Primitive(
        name="IdealInductor",
        desc="Ideal Inductor",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=IdealInductorParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["L", "Ind", "Inductor", "IdealL", "IdealInd"],
)


@paramclass
class PhysicalInductorParams:
    l = Param(dtype=Scalar, desc="Inductance (H)")


PhysicalInductor = _add(
    Primitive(
        name="PhysicalInductor",
        desc="Physical Inductor",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=PhysicalInductorParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["PhyL", "PhyInd", "IndPhy", "PhyInductor"],
)


ThreeTerminalInductor = _add(
    prim=Primitive(
        name="ThreeTerminalInductor",
        desc="Three Terminal Inductor",
        port_list=copy.deepcopy(ThreeTerminalPorts),
        paramtype=PhysicalInductorParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["Ind3", "PhyInd3", "IndPhy3", "PhyInductor3"],
)


@paramclass
class PhysicalShortParams:
    layer = Param(dtype=Optional[Scalar], desc="Metal layer", default=None)
    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)


PhysicalShort = _add(
    prim=Primitive(
        name="PhysicalShort",
        desc="Short-Circuit/ Net-Tie",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=PhysicalShortParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["Short"],
)


""" 
Sources
"""


@paramclass
class DcVoltageSourceParams:
    """`DcVoltageSource` Parameters"""

    dc = Param(dtype=Optional[Scalar], default=0, desc="DC Value (V)")
    ac = Param(dtype=Optional[Scalar], default=None, desc="AC Amplitude (V)")


DcVoltageSource = _add(
    prim=Primitive(
        name="DcVoltageSource",
        desc="DC Voltage Source",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=DcVoltageSourceParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=[
        "V",
        "Vdc",
        "Vsrc",
    ],
)


@paramclass
class PulseVoltageSourceParams:
    """`PulseVoltageSource` Parameters"""

    delay = Param(dtype=Optional[Scalar], default=None, desc="Time Delay (s)")
    v1 = Param(dtype=Optional[Scalar], default=None, desc="One Value (V)")
    v2 = Param(dtype=Optional[Scalar], default=None, desc="Zero Value (V)")
    period = Param(dtype=Optional[Scalar], default=None, desc="Period (s)")
    rise = Param(dtype=Optional[Scalar], default=None, desc="Rise time (s)")
    fall = Param(dtype=Optional[Scalar], default=None, desc="Fall time (s)")
    width = Param(dtype=Optional[Scalar], default=None, desc="Pulse width (s)")


PulseVoltageSource = _add(
    prim=Primitive(
        name="PulseVoltageSource",
        desc="Pulse Voltage Source",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=PulseVoltageSourceParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["Vpu", "Vpulse"],
)

Vpu = PulseVoltageSource
Vpulse = PulseVoltageSource


@paramclass
class SineVoltageSourceParams:
    """`SineVoltageSource` Parameters"""

    voff = Param(dtype=Optional[Scalar], default=None, desc="Offset (V)")
    vamp = Param(dtype=Optional[Scalar], default=None, desc="Amplitude (V)")
    freq = Param(dtype=Optional[Scalar], default=None, desc="Frequency (Hz)")
    td = Param(dtype=Optional[Scalar], default=None, desc="Delay (s)")
    phase = Param(dtype=Optional[Scalar], default=None, desc="Phase at td (degrees)")


SineVoltageSource = _add(
    prim=Primitive(
        name="SineVoltageSource",
        desc="Sine Voltage Source",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=SineVoltageSourceParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["Vsin"],
)


@paramclass
class CurrentSourceParams:
    dc = Param(dtype=Optional[Scalar], default=0, desc="DC Value (A)")


CurrentSource = _add(
    Primitive(
        name="CurrentSource",
        desc="Ideal DC Current Source",
        port_list=copy.deepcopy(PassivePorts),
        paramtype=CurrentSourceParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["I", "Idc", "Isrc"],
)


"""
Controlled Sources 
"""


@paramclass
class ControlledSourceParams:
    gain = Param(dtype=Scalar, default=1, desc="Gain in SI Units")


# Controlled Sources Port List
ControlledSourcePorts = [
    Port(name="p", desc="Output, Positive"),
    Port(name="n", desc="Output, Negative"),
    Port(name="cp", desc="Control, Positive"),
    Port(name="cn", desc="Control, Negative"),
]

VoltageControlledVoltageSource = _add(
    prim=Primitive(
        name="VoltageControlledVoltageSource",
        desc="Voltage Controlled Voltage Source",
        port_list=copy.deepcopy(ControlledSourcePorts),
        paramtype=ControlledSourceParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["Vcvs", "VCVS"],
)
CurrentControlledVoltageSource = _add(
    prim=Primitive(
        name="CurrentControlledVoltageSource",
        desc="Current Controlled Voltage Source",
        port_list=copy.deepcopy(ControlledSourcePorts),
        paramtype=ControlledSourceParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["Ccvs", "CCVS"],
)
VoltageControlledCurrentSource = _add(
    prim=Primitive(
        name="VoltageControlledCurrentSource",
        desc="Voltage Controlled Current Source",
        port_list=copy.deepcopy(ControlledSourcePorts),
        paramtype=ControlledSourceParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["Vccs", "VCCS"],
)
CurrentControlledCurrentSource = _add(
    prim=Primitive(
        name="CurrentControlledCurrentSource",
        desc="Current Controlled Current Source",
        port_list=copy.deepcopy(ControlledSourcePorts),
        paramtype=ControlledSourceParams,
        primtype=PrimitiveType.IDEAL,
    ),
    aliases=["Cccs", "CCCS"],
)

""" 
Bipolar Section 
"""


class BipolarType(Enum):
    """Bipolar Junction Transistor NPN/PNP Type Enumeration"""

    NPN = "NPN"
    PNP = "PNP"


@paramclass
class BipolarParams:
    """Bipolar Transistor Parameters"""

    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
    tp = Param(
        dtype=BipolarType, desc="Bipolar Type (NPN/ PNP)", default=BipolarType.NPN
    )
    model = Param(dtype=Optional[str], desc="Model (Name)", default=None)
    mult = Param(dtype=Optional[Scalar], desc="Multiplier", default=None)

    def __post_init__(self):
        """Value Checks"""
        if self.w is not None and self.w <= 0:
            raise ValueError(f"BipolarParams with invalid width {self.w}")
        if self.l is not None and self.l <= 0:
            raise ValueError(f"BipolarParams with invalid length {self.l}")


BipolarPorts = [Port(name="c"), Port(name="b"), Port(name="e")]

Bipolar = _add(
    prim=Primitive(
        name="Bipolar",
        desc="Bipolar Transistor",
        port_list=copy.deepcopy(BipolarPorts),
        paramtype=BipolarParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["Bjt", "BJT"],
)


def Npn(arg: Any = Default, **kwargs) -> Primitive:
    """Npn Constructor. A thin wrapper around `hdl21.primitives.Bipolar`"""
    bip = Bipolar(arg, **kwargs)
    bip.params = replace(bip.params, tp=BipolarType.NPN)
    return bip


def Pnp(arg: Any = Default, **kwargs) -> Primitive:
    """Pnp Constructor. A thin wrapper around `hdl21.primitives.Bipolar`"""
    bip = Bipolar(arg, **kwargs)
    bip.params = replace(bip.params, tp=BipolarType.PNP)
    return bip


""" 
Diodes
"""


@paramclass
class DiodeParams:
    w = Param(dtype=Optional[Scalar], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[Scalar], desc="Length in resolution units", default=None)
    model = Param(dtype=Optional[str], desc="Model (Name)", default=None)


Diode = _add(
    prim=Primitive(
        name="Diode",
        desc="Diode",
        # Despite not really being "passive", Diode does use the same `PassivePorts` list.
        port_list=copy.deepcopy(PassivePorts),
        paramtype=DiodeParams,
        primtype=PrimitiveType.PHYSICAL,
    ),
    aliases=["D"],
)



""" 
# Hdl21 Hardware Description Library 
"""

__version__ = "7.0.0"  # VLSIR_VERSION

# Internal (python) module aliases, overridden by names such as the `module` decorator function.
from . import module as _module_module
from . import bundle as _bundle_module
from . import external_module as _external_module_module
from . import instance as _instance_module
from . import generator as _generator_module

from . import params
from .params import *

from .instance import *
from .signal import *
from .slice import *
from .concat import *
from .noconn import *
from .external_module import *
from .module import *
from .generator import *
from .bundle import *
from .role import *
from .netlisting import *
from .instantiable import *
from .diff_pair import *
from .props import Properties
from .scalar import Scalar
from .literal import Literal

from . import primitives
from .primitives import *

from . import prefix
from .prefix import Prefix, Prefixed

# Import these as modules, but not their contents
from . import generators
from . import sim
from . import pdk
from . import proto
from .proto import to_proto, from_proto

from .walker import HierarchyWalker

from . import elab
from .elab import *

# Update all the forward type-references throughout our many `@datatype`s
from .datatype import _update_forward_refs

_update_forward_refs()



"""
# Width Resolution Helper 

Function to determine the `width` of connectable types. 
"""

# Std-Lib Imports
from typing import Union, Callable

# Local imports
from ...connect import Connectable
from ...portref import PortRef
from ...signal import Signal
from ...slice import Slice
from ...concat import Concat
from ...noconn import NoConn
from ...bundle import (
    AnonymousBundle,
    BundleInstance,
    BundleRef,
)
from ..helpers.resolve_ref_types import resolve_bundleref_type, resolve_portref_type

# HasWidth & Sliceable Type Uhions
# Everything that has, or at least sometimes has, an integer Signal-style `width`.
# AND everything that can be `Slice`d.
# These happen to be the same set of types, but can be broken up
# if the two should ever differ.
Sliceable = HasWidth = Union[Signal, Slice, Concat, BundleRef, PortRef]


def fail(msg: str) -> None:
    # The default failure handler
    raise RuntimeError(msg)


def width(conn: Connectable, failer: Callable = fail) -> int:
    """Get the `width` of a conn. Largely dispatches across conn types.
    Optional function-valued argument `failer` is passed all errors.
    This is commonly used to pass failure information and control back to `ElabPass`s."""

    # A reminder, as of this writing:
    # Connectable = Union["Signal", "Slice", "Concat", "NoConn", "PortRef", "BundleInstance", "AnonymousBundle", "BundleRef"]
    # (Man that Rust-style enum would be nice)

    # Invalid cases
    if isinstance(conn, (BundleInstance, AnonymousBundle, NoConn)):
        return failer(f"Invalid `width` of {conn}")

    # Typical "scalar" cases
    if isinstance(conn, Signal):
        return conn.width
    if isinstance(conn, Slice):
        return conn.width
    if isinstance(conn, Concat):
        return sum([width(p) for p in conn.parts])

    # References
    if isinstance(conn, (BundleRef, PortRef)):
        return ref_width(conn, failer)

    return failer(f"Invalid `width` of {conn}")


def ref_width(ref: Union[PortRef, BundleRef], failer: Callable = fail) -> int:
    """Get a reference's width from its referent.
    Fail if this resolves to a Bundle Instance.
    And cache the result on the Ref for future use.

    Optional function-valued argument `failer` is passed all errors.
    This is commonly used to pass failure information and control back to `ElabPass`s."""

    if ref._width is not None:
        return ref._width

    # Get the referent
    if isinstance(ref, BundleRef):
        referent = resolve_bundleref_type(ref, failer)
    elif isinstance(ref, PortRef):
        referent = resolve_portref_type(ref)
    else:
        return failer(f"Invalid arg to ref_width {ref}")

    # And get its width
    if isinstance(referent, Signal):
        ref._width = width(referent, failer)
        return ref._width
    if isinstance(referent, BundleInstance):
        return failer(f"Invalid `width` of Bundle {referent}")
    return failer(f"Invalid `width` of {referent}")



from .corner import Corner, CornerType, CmosCorner


def test_corner1():
    assert str(Corner.TYP) == "Corner.TYP"
    assert str(Corner.FAST) == "Corner.FAST"
    assert str(Corner.SLOW) == "Corner.SLOW"


def test_cornertype1():
    assert str(CornerType.MOS) == "CornerType.MOS"
    assert str(CornerType.CMOS) == "CornerType.CMOS"
    assert str(CornerType.RES) == "CornerType.RES"
    assert str(CornerType.CAP) == "CornerType.CAP"


def test_cmos1():
    assert str(CmosCorner.TT) == "CmosCorner.TT"
    assert str(CmosCorner.FF) == "CmosCorner.FF"
    assert str(CmosCorner.SS) == "CmosCorner.SS"
    assert str(CmosCorner.FS) == "CmosCorner.FS"
    assert str(CmosCorner.SF) == "CmosCorner.SF"



"""
# Hdl21 Unit Tests
Pulling directly from the documentation snippets, and making sure they work as claimed. 
"""

import hdl21 as _h


@_h.module
class AnotherModule:
    # This module is not defined in the docs, but of course is needed to actually run.
    a, b, c = _h.Ports(3)


def test_module1():
    import hdl21 as h

    m = h.Module(name="MyModule")
    m.i = h.Input()
    m.o = h.Output(width=8)
    m.s = h.Signal()
    m.a = AnotherModule()


def test_module2():
    import hdl21 as h

    @h.module
    class MyModule:
        i = h.Input()
        o = h.Output(width=8)
        s = h.Signal()
        a = AnotherModule()


def test_module3():
    import hdl21 as h

    @h.module
    class MyModule:
        a, b = h.Inputs(2)
        c, d, e = h.Outputs(3, width=16)
        z, y, x, w = h.Signals(4)


def test_module4():
    import hdl21 as h

    # Create a module
    m = h.Module()
    # Create its internal Signals
    m.a, m.b, m.c = h.Signals(3)
    # Create an Instance
    m.i1 = AnotherModule()
    # And wire them up
    m.i1.a = m.a
    m.i1.b = m.b
    m.i1.c = m.c


def test_module4():
    import hdl21 as h

    # Create a module
    m = h.Module()
    # Create the Instances
    m.i1 = AnotherModule()
    m.i2 = AnotherModule()
    # Call one to connect them
    m.i1(a=m.i2.a, b=m.i2.b, c=m.i2.c)


def test_module5():
    import hdl21 as h

    # Create a module
    m = h.Module()
    # Create the Instance `i1`
    m.i1 = AnotherModule()
    # Create another Instance `i2`, and connect to `i1`
    m.i2 = AnotherModule(a=m.i1.a, b=m.i1.b, c=m.i1.c)


@_h.paramclass
class MyParams:
    # Reused by the generator tests to follow
    w = _h.Param(dtype=int, default=8, desc="Width")


def test_generator1():
    import hdl21 as h
    from hdl21.generator import GeneratorCall

    @h.generator
    def MyFirstGenerator(params: MyParams) -> h.Module:
        # A very exciting first generator function
        m = h.Module()
        m.i = h.Input(width=params.w)
        return m

    m = MyFirstGenerator(MyParams(w=16))
    call = m._generated_by
    assert isinstance(call, GeneratorCall)
    assert call.params == MyParams(w=16)


def test_generator2():
    import hdl21 as h

    @h.generator
    def MySecondGenerator(params: MyParams) -> h.Module:
        # A very exciting (second) generator function
        @h.module
        class MySecondGen:
            i = h.Input(width=params.w)

        return MySecondGen

    # FIXME: checks


def test_generator3():
    import hdl21 as h

    @h.generator
    def MyThirdGenerator(params: MyParams) -> h.Module:
        # Create an internal Module
        @h.module
        class Inner:
            i = h.Input(width=params.w)

        # Manipulate it a bit
        Inner.o = h.Output(width=2 * Inner.i.width)

        # Instantiate that in another Module
        @h.module
        class Outer:
            inner = Inner()

        # And manipulate that some more too
        Outer.inp = h.Input(width=params.w)
        return Outer

    # FIXME: checks


def test_params1():
    import hdl21 as h

    @h.paramclass
    class MyParams:
        # Required
        width = h.Param(dtype=int, desc="Width. Required")
        # Optional - including a default value
        text = h.Param(dtype=str, desc="Optional string", default="My Favorite Module")

    p = MyParams(width=8, text="Your Favorite Module")
    assert p.width == 8  # Passes. Note this is an `int`, not a `Param`
    assert p.text == "Your Favorite Module"  # Also passes


def test_params2():
    import hdl21 as h
    from dataclasses import asdict

    @h.paramclass
    class Inner:
        i = h.Param(dtype=int, desc="Inner int-field")

    @h.paramclass
    class Outer:
        inner = h.Param(dtype=Inner, desc="Inner fields")
        f = h.Param(dtype=float, desc="A float", default=3.14159)

    # Create from a (nested) dictionary literal
    d1 = {"inner": {"i": 11}, "f": 22.2}
    o = Outer(**d1)
    # Convert back to another dictionary
    d2 = asdict(o)
    # And check they line up
    assert d1 == d2


def test_inline_params_construction():
    import hdl21 as h

    @h.paramclass
    class MyParams:
        width = h.Param(dtype=int, desc="Width. Required")
        text = h.Param(dtype=str, desc="Optional string", default="My Favorite Module")

    @h.generator
    def MyGen(params: MyParams) -> h.Module:
        ...
        # Module content is really irrelevant; really just shows caching
        return h.Module()

    assert MyGen(width=8, text="My Favorite Module") == MyGen(
        MyParams(width=8, text="My Favorite Module")
    )


def test_external_module_primitive_example():
    """Test for the Primitives and ExternalModules example"""

    import hdl21 as h
    from hdl21.prefix import µ
    from hdl21.primitives import Diode

    @h.paramclass
    class BandGapParams:
        self_destruct = h.Param(
            dtype=bool,
            desc="Whether to include the self-destruction feature",
            default=True,
        )

    BandGap = h.ExternalModule(
        name="BandGap",
        desc="Example ExternalModule, defined outside Hdl21",
        port_list=[h.Port(name="vref"), h.Port(name="enable")],
        paramtype=BandGapParams,
    )

    params = BandGapParams(self_destruct=False)  # Watch out there!

    @h.module
    class BandGapPlus:
        vref, enable = h.Signals(2)
        # Instantiate the `ExternalModule` defined above
        bg = BandGap(params)(vref=vref, enable=enable)
        # ...Anything else...

    @h.module
    class DiodePlus:
        p, n = h.Signals(2)
        # Parameterize, instantiate, and connect a `primitives.Diode`
        d = Diode(w=1 * µ, l=1 * µ)(p=p, n=n)
        # ... Everything else ...

    """ 
    The part outside the docs: testing it 
    """

    assert isinstance(BandGapPlus, h.Module)
    assert isinstance(BandGapPlus.bg, h.Instance)
    assert isinstance(BandGapPlus.bg.of, h.ExternalModuleCall)
    assert isinstance(DiodePlus, h.Module)
    assert isinstance(DiodePlus.d, h.Instance)
    assert isinstance(DiodePlus.d.of, h.PrimitiveCall)


def test_scalar():
    """Test the `Scalar` example"""
    import hdl21 as h
    from hdl21.prefix import NANO, µ
    from decimal import Decimal

    @h.paramclass
    class MyMosParams:
        w = h.Param(dtype=h.Scalar, desc="Width", default=1e-6)
        l = h.Param(dtype=h.Scalar, desc="Length", default="w/5")

    # Example instantiations
    p = MyMosParams()  # Default values
    assert p.w == h.scalar.to_scalar(Decimal(1e-6))
    assert p.l == h.scalar.to_scalar("w/5")
    assert isinstance(p.w, h.Prefixed)
    assert isinstance(p.l, h.Literal)

    p = MyMosParams(w=Decimal(1e-6), l=3 * µ)
    assert isinstance(p.w, h.Prefixed)
    assert isinstance(p.l, h.Prefixed)
    assert p.w == h.scalar.to_scalar(Decimal(1e-6))
    assert p.l == h.scalar.to_scalar(3 * µ)

    p = MyMosParams(w=h.Literal("sim_param_width"), l=h.Prefixed.new(20, NANO))
    assert isinstance(p.w, h.Literal)
    assert isinstance(p.l, h.Prefixed)
    assert p.w == h.scalar.to_scalar(h.Literal("sim_param_width"))
    assert p.l == h.scalar.to_scalar(h.Prefixed.new(20, NANO))

    p = MyMosParams(w="11*l", l=11)
    assert isinstance(p.w, h.Literal)
    assert isinstance(p.l, h.Prefixed)
    assert p.w == h.scalar.to_scalar("11*l")
    assert p.l == h.scalar.to_scalar(11)



"""
# PortRef Resolution 

Creates concrete `Signal`s and `BundleInstance`s to replace `PortRef`s.
"""

# Std-Lib Imports
import copy
from typing import Union, List, Optional, Dict

# Local imports
from ...connect import Connectable
from ...instance import _get_connref
from ...instantiable import (
    io,
    Instantiable,
    ExternalModuleCall,
    PrimitiveCall,
)
from ...module import Module
from ...portref import PortRef
from ...bundle import BundleInstance, BundleRef, AnonymousBundle
from ...signal import PortDir, Signal, Visibility
from ...noconn import NoConn
from ..helpers.resolve_ref_types import update_ref_deps

# Import the base class
from .base import ElabPass

# Union of the types which serve as "source signals",
# i.e. the things which we are resolve `PortRef`s *to*.
# If we find one of these connected to a group of connected ports,
# it becomes the replacement connection for all of them.
Source = Union[Signal, BundleInstance, BundleRef, AnonymousBundle]

# Union of the types which can serve as (generalized) Ports:
# either Signals or Bundle Instances
PortType = Union[Signal, BundleInstance]


class ResolvePortRefs(ElabPass):
    """Resolve all `PortRef`s on all `Instance`s.

    This pass resolves "Instance to Instance" connections such as:
    ```
    inst1 = Module1()
    inst2 = Module2(port2=inst1.port1)
    ```

    After this pass, all such connections resolve to concrete
    connectable objects, all of which are members of the `Source` type-union
    defined above.

    While not terribly consistent with its name, this pass also resolves
    all `NoConn` connections into concrete `Signal`s, or fails if any `NoConn`s
    are invalidly connected between one another.
    """

    def elaborate_module(self, module: Module) -> Module:
        """
        Resolve and replace all Instance `PortRef`s in `module`.

        In the process, all `Connectable`s are annotated with a set of `_connected_ports`.
        These `_connected_ports` are relied upon by later elaboration stages,
        and must be maintained hereafter, e.g. when making connection updates.
        """

        instancelike = (
            list(module.instances.values())
            + list(module.instarrays.values())
            + list(module.instbundles.values())
        )

        # Collect up all `PortRef`s for all instances in the module
        # FIXME: move from SetList to a regular Set. Thus far breaks one test, somehow.
        module_portrefs = SetList()
        for inst in instancelike:
            # Populate the module-level set of PortRefs
            for portref in inst._refs.portrefs.values():
                module_portrefs.add(portref)

            # FIXME: add the `NoConn`s here, although it's not clear we *really* need these checks on them
            for portname, conn in inst.conns.items():
                if isinstance(conn, NoConn):
                    module_portrefs.add(_get_connref(inst, portname))

        def follow(pref: PortRef, group: SetList) -> None:
            """Closure to recursively follow `pref`, adding its outward and inward connections to `group`.
            Removes encountered entries from `module_portrefs` along the way."""

            if pref in group:
                return  # Already done
            if pref in module_portrefs:
                module_portrefs.remove(pref)

            # Add the portref to the group
            group.add(pref)

            # Add, and if necessary recursively follow, its instance connection
            conn = pref.inst.conns.get(pref.portname, None)
            if isinstance(conn, PortRef):
                follow(conn, group)
            else:  # Add ultimate signal `Source`s to the group
                group.add(conn)

            # And recursively follow its connected ports
            for connected_port in pref._connected_ports:
                follow(connected_port, group)

        # Collect groups of connected `PortRef`s
        groups: List[List[Optional[Connectable]]] = list()

        while module_portrefs:
            # Create a new group, and recursively follow port refs to populate it
            group = SetList()
            follow(module_portrefs.pop(), group)

            # Filter out the `None` entries before sticking in the `groups` list-of-lists.
            group: List[Connectable] = [x for x in group.order if x is not None]
            groups.append(group)

        # For each group, find and/or create a Signal to replace all the PortRefs with.
        for group in groups:
            self.handle_group(module, group)

        return module

    def handle_group(self, module: Module, group: List[Connectable]) -> None:
        """Handle a `group` worth of connected Ports and NoConns"""

        # First cover `NoConn` connections, checking for any invalid multiple-conns to them.
        if any([isinstance(n, NoConn) for n in group]):
            return self.handle_noconn(module, group)
        return self.handle_portconn(module, group)

    def handle_portconn(self, module: Module, group: List[Connectable]):
        """Handle re-connecting a list of connected `PortRef`s.
        Creates and adds a fresh new `Signal` if one does not already exist."""

        group_port_refs = [x for x in group if isinstance(x, PortRef)]

        # Find any existing, declared `Source` connected to `group`.
        # And if we don't find any, go about naming and creating one.
        source: Optional[Source] = self.find_source(group)
        if source is None:
            source = self.create_source(module, group_port_refs)

        # Resolve each PortRef with `source` as its referent,
        # and reconnect it everywhere it's connected.
        for portref in group_port_refs:
            resolve_portref(portref, source)

    def find_source(self, group: List[Connectable]) -> Optional[Source]:
        """Find any existing, declared `Source` connected to `group`.
        Returns None is no `Source`s are connected to any element in the group."""

        # Extract which of the `group` are connected to `Source`s
        sources: List[Source] = [s for s in group if isinstance(s, Source.__args__)]

        # Three relevant cases then emerge:
        # * (a) Zero sources. No problem, create one.
        # * (b) A single source. Also good; this will be connected to all other ports in the group.
        # * (c) More than one source.
        #   * This *shouldn't* be possible. If it is possible, we haven't figured out how.
        #   * This method raises a `RuntimeError` if this somehow happens.

        if len(sources) == 0:
            return None
        if len(sources) == 1:
            return sources[0]

        # More than one source, somehow. Error time.
        msg = f"Internal error: invalid connection, with multiple Source-Signals {sources}, "
        msg += f"shorting Ports {[(p.inst.name, p.portname) for p in group]}"
        self.fail(msg)

    def which_portref_to_name(self, group: List[PortRef]) -> PortRef:
        """Figure out which PortRef in `group` to use for its new signal name.
        This generally follows "the port the other ones refer to", i.e. for:
        ```
        i1.p = i0.p
        i2.p = i0.p
        i3.p = i1.p
        i4.p = i3.p
        ```
        The Instance `i0` has the sole None-connected Port `p`, and gets "naming rights".

        In other cases there is no such instance, such as:
        ```
        i0.p = i1.p
        i1.p = i0.p # Connected back!
        ```
        In these cases, the first Instance name in alphabetical order is used.
        """

        # Sort out which PortRefs have no connection
        conn = lambda p: p.inst.conns.get(p.portname, None)
        connected_to_none = [p for p in group if conn(p) is None]

        if len(connected_to_none) == 1:
            return connected_to_none[0]

        if len(connected_to_none) > 1:
            self.fail(f"Invalid PortRef group: {group}")

        # Nothing "unconnected". Find the instance one with the lowest (alphabetical) name.
        ordered = sorted(group, key=lambda p: p.inst.name)
        return ordered[0]

    def create_source(self, module: Module, group: List[PortRef]) -> PortType:
        """Create a new `Signal`, parametrized and named to connect to the `PortRef`s in `group`."""
        from ...instantiable import io

        # Sort out which PortRef to use for the new signal name.
        portref = self.which_portref_to_name(group)

        # Get its Module for its IO
        # Note if the other entries in `group` have incompatible IO, this will be flagged by a later pass.
        ios = io_for_resolving(portref.inst.of)

        port = ios.get(portref.portname, None)
        if port is None:  # Clone it, and remove any Port-attributes
            msg = f"Invalid port `{portref.portname}` on Instance `{portref.inst.name}` in Module `{module.name}`"
            self.fail(msg)

        # Copy that port into an internal Signal / Bundle
        sig = self.copy_port(port)

        # Rename it and add it to the Module namespace
        signame = self.flatname(
            segments=[f"{portref.inst.name}_{portref.portname}"],
            avoid=module.namespace,
        )
        sig.name = signame
        module.add(sig)
        return sig

    def copy_port(self, port: PortType) -> PortType:
        """Copy a port into an internal Signal or BundleInstance"""

        if isinstance(port, Signal):
            # Make a copy, and update its port-level visibility to internal
            sig = copy.copy(port)
            sig.vis = Visibility.INTERNAL
            sig.direction = PortDir.NONE
            return sig

        if isinstance(port, BundleInstance):
            # Copy the bundle, removing port and role fields.
            return BundleInstance(of=port.of, port=False, role=None)

        self.fail(f"Invalid Port Type `{port}`")

    def handle_noconn(self, module: Module, group: List[Connectable]):
        """Handle a group with a `NoConn`."""
        # First check for validity of the group, i.e. that the `NoConn` only connects to *one* port.
        if len(group) > 2:
            msg = f"Invalid multiply-connected `NoConn`, including {group} in {module}"
            self.fail(msg)
        # So `group` has two entries: a `NoConn` and a `PortRef`
        if isinstance(group[0], NoConn):
            return self.replace_noconn(module, portref=group[1], noconn=group[0])
        return self.replace_noconn(module, portref=group[0], noconn=group[1])

    def replace_noconn(self, module: Module, portref: PortRef, noconn: NoConn):
        """Replace `noconn` with a newly minted `Signal` or `BundleInstance`."""

        # Get the target Module's port-object corresponding to `portref`
        io_dict: Dict[str, Connectable] = io_for_resolving(portref.inst.of)
        port = io_dict.get(portref.portname, None)
        if port is None:
            msg = f"Invalid port connection to `{portref}` in Module `{module}`"
            self.fail(msg)

        # Copy any relevant attributes of the Port
        sig = self.copy_port(port)

        # Set the signal name, either from the NoConn or the instance/port names
        if noconn.name is not None:
            sig.name = noconn.name
        else:
            sig.name = self.flatname(
                segments=[f"{portref.inst.name}_{portref.portname}"],
                avoid=module.namespace,
            )

        # Add the new signal, and connect it to `inst`
        module.add(sig)
        portref.inst.connect(portref.portname, sig)


class SetList:
    """A common combination of a hash-set and ordered list of the same items.
    Used for keeping ordered items while maintaining quick membership testing.

    FIXME: this is really using a LIST for now, slowing down membership tests,
    but should be set-based when some connectable types, particularly `Signal`, are ready for it.
    """

    def __init__(self):
        # self.set = set()
        self.list = list()

    def __bool__(self):
        return bool(self.list)

    def __contains__(self, item):
        # return item in self.set
        return item in self.list

    def add(self, item):
        # if item not in self.set:
        if item not in self.list:
            # self.set.add(item)
            self.list.append(item)

    def remove(self, item):
        # if item not in self.set:
        if item not in self.list:
            raise RuntimeError
            # self.set.add(item)
        self.list.remove(item)

    def pop(self):
        # if item not in self.set:
        if not self.list:
            raise RuntimeError
            # self.set.add(item)
        return self.list.pop(0)

    @property
    def order(self):
        return self.list


def resolve_portref(pref: PortRef, to: Connectable) -> None:
    """# Resolve a `PortRef` to its referent `Connectable`."""

    if pref.resolved is to:
        return  # Already resolved
    if pref.resolved is not None:
        raise ValueError(f"PortRef {pref} already resolved")

    # Give the `PortRef` a reference to its referent
    pref.resolved = to

    # Connect the "primary" instance to the referent
    # Note this differs between `PortRef` and `BundleRef`;
    # the latter has no "primary instance" to connect to.
    pref.inst.connect(pref.portname, to)

    # Update all downstream dependencies, e.g. connected ports, slices
    update_ref_deps(pref, to)


def io_for_resolving(i: Instantiable) -> Dict[str, "Connectable"]:
    """Get the relevant IOs of Instantiable `i` for port-ref resolution.
    Depending on the elaboration state of `i`, this is either the "bundled" or "flattened" IOs.
    """

    if isinstance(i, (ExternalModuleCall, PrimitiveCall)):
        # These do not have Bundle-valued ports
        return copy.copy(i.ports)

    if not isinstance(i, Module):
        raise TypeError(f"Invalid Instantiable: {i}")

    if i._pre_flattening_io is not None:
        return copy.copy(i._pre_flattening_io)
    return io(i)



"""
# Hdl21 Unit Tests

Not necessarily exclusive to `Bundle`s, but focusing on them.
"""
import copy, pytest
from enum import Enum, auto

import hdl21 as h


def test_bundle1():
    # Create an bundle

    i1 = h.Bundle(name="MyFirstBundle")
    i1.s1 = h.Signal()
    i1.s2 = h.Signal()

    assert isinstance(i1, h.Bundle)
    assert isinstance(i1.s1, h.Signal)
    assert isinstance(i1.s2, h.Signal)

    ii1 = i1()
    assert isinstance(ii1, h.BundleInstance)
    assert ii1.role is None
    assert ii1.port == False


def test_bundle2():
    # Wire up a few Modules via bundles

    MySecondBundle = h.Bundle(name="MySecondBundle")
    MySecondBundle.s = h.Signal()

    m1 = h.Module(name="M1")
    m1.i = MySecondBundle(port=True)
    m2 = h.Module(name="M2")
    m2.i = MySecondBundle(port=True)

    # Now create a parent Module connecting the two
    m3 = h.Module(name="M3")
    m3.i1 = m1()
    m3.i2 = m2(i=m3.i1.i)
    assert "i1_i" not in m3.namespace

    # First run the "implicit bundles" pass, and see that an explicit one is created
    from hdl21.elab import Elaborator, set_elaborator, reset_elaborator
    from hdl21.elab.passes import ResolvePortRefs

    e = Elaborator(passes=[ResolvePortRefs])
    set_elaborator(e)

    m3 = e.elaborate(m3)
    assert isinstance(m3, h.Module)
    assert isinstance(m3.i1, h.Instance)
    assert isinstance(m3.i2, h.Instance)
    assert "i1_i" in m3.namespace
    reset_elaborator()

    # Now elaborate it the rest of the way, to scalar signals
    m3 = h.elaborate(m3)
    assert "i1_i" not in m3.namespace
    assert "i1_i_s" in m3.namespace
    assert isinstance(m3.get("i1_i_s"), h.Signal)
    assert m3.get("i1_i_s") in m3.i1.conns.values()


def test_bundle3():
    # Test the bundle-definition decorator

    @h.bundle
    class Diff:  # Differential Signal Bundle
        p = h.Signal()
        n = h.Signal()

    @h.bundle
    class DisplayPort:  # DisplayPort, kinda
        main_link = Diff()
        aux = h.Signal()

    assert isinstance(DisplayPort, h.Bundle)
    assert isinstance(DisplayPort(), h.BundleInstance)
    assert isinstance(DisplayPort.main_link, h.BundleInstance)
    assert isinstance(DisplayPort.main_link.p, h.BundleRef)
    assert isinstance(DisplayPort.main_link.n, h.BundleRef)
    assert isinstance(DisplayPort.aux, h.Signal)
    assert isinstance(Diff, h.Bundle)
    assert isinstance(Diff(), h.BundleInstance)
    assert isinstance(Diff.p, h.Signal)
    assert isinstance(Diff.n, h.Signal)

    # Instantiate one in a Module
    m = h.Module(name="M")
    m.dp = DisplayPort()
    assert isinstance(m.dp, h.BundleInstance)
    assert len(m.bundles) == 1
    assert len(m.signals) == 0

    # And elaborate it
    h.elaborate(m)

    assert not hasattr(m, "dp")
    assert len(m.bundles) == 0
    assert len(m.signals) == 3
    assert isinstance(m.get("dp_aux"), h.Signal)
    assert isinstance(m.get("dp_main_link_p"), h.Signal)
    assert isinstance(m.get("dp_main_link_n"), h.Signal)


def test_bundle4():
    # Test bundle roles

    @h.bundle
    class Diff:  # Differential Signal Bundle
        p = h.Signal()
        n = h.Signal()

    @h.bundle
    class HasRoles:  # An Bundle with Roles
        class Roles(Enum):  # USB-Style Role Nomenclature
            HOST = auto()
            DEVICE = auto()

        Roles = h.RoleSet.from_enum(Roles)

        # Create signals going in either direction
        tx = h.Signal(src=Roles.HOST, dest=Roles.DEVICE)
        rx = h.Signal(src=Roles.DEVICE, dest=Roles.HOST)

        # And create differential versions thereof
        txd = Diff(src=Roles.HOST, dest=Roles.DEVICE)
        rxd = Diff(src=Roles.DEVICE, dest=Roles.HOST)

    hr = HasRoles()
    assert isinstance(HasRoles, h.Bundle)
    assert isinstance(HasRoles.roles, h.RoleSet)
    assert isinstance(HasRoles.Roles, h.RoleSet)
    assert isinstance(hr, h.BundleInstance)
    assert isinstance(HasRoles.tx, h.Signal)
    assert isinstance(HasRoles.rx, h.Signal)
    assert isinstance(HasRoles.txd, h.BundleInstance)
    assert isinstance(HasRoles.rxd, h.BundleInstance)

    @h.module
    class Host:
        # A thing with a HOST-roled bundle-port
        hr = HasRoles(port=True, role=HasRoles.Roles.HOST)

    @h.module
    class Device:
        # A thing with a DEVICE-roled bundle-port
        hr = HasRoles(port=True, role=HasRoles.Roles.DEVICE)

    @h.module
    class System:
        # Parent system-module including a host and device
        host = Host()
        devc = Device(hr=host.hr)

    assert isinstance(System, h.Module)
    assert isinstance(System.host, h.Instance)
    assert isinstance(System.devc, h.Instance)
    assert "host_hr" not in System.namespace

    # First run the "implicit bundles" pass, and see that an explicit one is created
    from hdl21.elab import Elaborator
    from hdl21.elab.passes import ResolvePortRefs

    e = Elaborator(passes=[ResolvePortRefs])

    sys = e.elaborate(System)
    assert "host_hr" in sys.namespace

    # Now expand the rest of the way, down to scalar signals
    # Check that bundle went away, and its constituent signals replaced it
    sys = h.elaborate(sys)
    assert "host_hr" not in sys.namespace
    assert "host_hr_tx" in sys.namespace
    assert "host_hr_rx" in sys.namespace
    assert "host_hr_txd_p" in sys.namespace
    assert "host_hr_txd_n" in sys.namespace
    assert "host_hr_rxd_p" in sys.namespace
    assert "host_hr_rxd_n" in sys.namespace


def test_bigger_bundles():
    """Test a slightly more elaborate Bundle-based system"""

    @h.roleset
    class HostDevice(Enum):
        # A pair of roles `Host` and `Device`, as commonly used by USB, JTAG, and similar buses.
        HOST = auto()
        DEVICE = auto()

    @h.bundle
    class Jtag:
        # Jtag Bundle
        roles = HostDevice
        tck, tdi, tms = h.Signals(3, src=roles.HOST, dest=roles.DEVICE)
        tdo = h.Signal(src=roles.DEVICE, dest=roles.HOST)

    @h.bundle
    class Uart:
        # Uart Bundle
        # Note the UART bundle is not role-based; `tx` is always output, `rx` is always input.
        tx = h.Output()
        rx = h.Input()

    @h.bundle
    class Spi:
        # Spi Bundle
        roles = HostDevice
        sck, cs = h.Signals(2, src=roles.HOST, dest=roles.DEVICE)
        dq = h.Signal(src=roles.DEVICE, dest=roles.HOST, width=4)

    @h.module
    class Chip:
        spi = Spi(role=HostDevice.HOST, port=True)
        jtag = Jtag(role=HostDevice.DEVICE, port=True)
        uart = Uart(port=True)
        ...  # Actual internal content, which likely connects these down *many* levels of hierarchy

    @h.module
    class SpiFlash:
        # A typical flash memory with a SPI port
        spi = Spi(role=HostDevice.DEVICE, port=True)

    @h.module
    class Board:
        # A typical embedded board, featuring a custom chip, SPI-connected flash, and JTAG port
        jtag = Jtag(role=HostDevice.DEVICE, port=True)
        uart = Uart(port=True)

        chip = Chip(jtag=jtag, uart=uart)
        flash = SpiFlash(spi=chip.spi)

    @h.module
    class Tester:
        # A typical test-widget with a JTAG port
        jtag = Jtag(role=HostDevice.HOST, port=True)
        uart = Uart(port=True)

    @h.module
    class TestSystem:
        # A system in which `Tester` can test `Board`
        jtag = Jtag()
        u0, u1 = h.Signals(2)

        # Connect the two, swapping UART `rx` and `tx`
        board = Board(jtag=jtag, uart=h.bundlize(tx=u0, rx=u1))
        tester = Tester(jtag=jtag, uart=h.bundlize(tx=u1, rx=u0))

    assert isinstance(TestSystem.jtag, h.BundleInstance)
    assert isinstance(TestSystem.tester, h.Instance)
    assert isinstance(TestSystem.board, h.Instance)

    assert isinstance(TestSystem.tester.uart, h.PortRef)
    assert isinstance(TestSystem.board.uart, h.PortRef)

    # Run this through elaboration
    h.elaborate(TestSystem)

    # Post-elab checks
    # TestSystem
    assert isinstance(TestSystem.tester, h.Instance)
    assert isinstance(TestSystem.board, h.Instance)
    assert TestSystem.tester.of is Tester
    assert TestSystem.board.of is Board
    assert not hasattr(TestSystem, "jtag")
    assert not hasattr(TestSystem, "uart")
    assert "u0" in TestSystem.namespace
    assert "u1" in TestSystem.namespace
    assert len(TestSystem.ports) == 0
    assert len(TestSystem.signals) == 6
    assert len(TestSystem.instances) == 2
    assert isinstance(TestSystem.get("u0"), h.Signal)
    assert isinstance(TestSystem.get("u1"), h.Signal)
    assert isinstance(TestSystem.get("jtag_tck"), h.Signal)
    assert isinstance(TestSystem.get("jtag_tdi"), h.Signal)
    assert isinstance(TestSystem.get("jtag_tdo"), h.Signal)
    assert isinstance(TestSystem.get("jtag_tms"), h.Signal)

    # Tester
    assert len(Tester.ports) == 6
    assert len(Tester.signals) == 0
    assert len(Tester.instances) == 0
    assert isinstance(Tester.get("jtag_tck"), h.Signal)
    assert Tester.get("jtag_tck").vis == h.signal.Visibility.PORT
    assert Tester.get("jtag_tdo").vis == h.signal.Visibility.PORT
    assert Tester.get("jtag_tdi").vis == h.signal.Visibility.PORT
    assert Tester.get("jtag_tms").vis == h.signal.Visibility.PORT
    assert Tester.get("uart_tx").vis == h.signal.Visibility.PORT
    assert Tester.get("uart_rx").vis == h.signal.Visibility.PORT

    # Board
    assert len(Board.ports) == 6
    assert len(Board.signals) == 3  # SPI signals
    assert len(Board.instances) == 2
    assert Board.chip.of is Chip
    assert Board.flash.of is SpiFlash
    assert Board.get("jtag_tck").vis == h.signal.Visibility.PORT
    assert Board.get("jtag_tdo").vis == h.signal.Visibility.PORT
    assert Board.get("jtag_tdi").vis == h.signal.Visibility.PORT
    assert Board.get("jtag_tms").vis == h.signal.Visibility.PORT
    assert Board.get("uart_tx").vis == h.signal.Visibility.PORT
    assert Board.get("uart_rx").vis == h.signal.Visibility.PORT
    assert Board.get("chip_spi_sck").vis == h.signal.Visibility.INTERNAL
    assert Board.get("chip_spi_cs").vis == h.signal.Visibility.INTERNAL
    assert Board.get("chip_spi_dq").vis == h.signal.Visibility.INTERNAL

    # Chip
    assert len(Chip.ports) == 9
    assert len(Chip.signals) == 0
    assert len(Chip.instances) == 0
    assert Chip.get("jtag_tck").vis == h.signal.Visibility.PORT
    assert Chip.get("jtag_tdo").vis == h.signal.Visibility.PORT
    assert Chip.get("jtag_tdi").vis == h.signal.Visibility.PORT
    assert Chip.get("jtag_tms").vis == h.signal.Visibility.PORT
    assert Chip.get("spi_sck").vis == h.signal.Visibility.PORT
    assert Chip.get("spi_cs").vis == h.signal.Visibility.PORT
    assert Chip.get("spi_dq").vis == h.signal.Visibility.PORT
    assert Chip.get("uart_tx").vis == h.signal.Visibility.PORT
    assert Chip.get("uart_rx").vis == h.signal.Visibility.PORT

    # SpiFlash
    assert len(SpiFlash.ports) == 3
    assert len(SpiFlash.signals) == 0
    assert len(SpiFlash.instances) == 0
    assert SpiFlash.get("spi_sck").vis == h.signal.Visibility.PORT
    assert SpiFlash.get("spi_cs").vis == h.signal.Visibility.PORT
    assert SpiFlash.get("spi_dq").vis == h.signal.Visibility.PORT


def test_bad_bundle_conn():
    """Test invalid Bundle connections"""

    @h.bundle
    class P:
        p = h.Signal(width=3)

    @h.bundle
    class R:
        z = h.Signal(width=11)

    @h.module
    class C:
        p = P(port=True)  # `P`-type Bundle

    @h.module
    class Q:
        r = R()  # `R`-type Bundle
        c = C(p=r)  # <= Bad connection here

    with pytest.raises(RuntimeError):
        h.elaborate(Q)


def test_copy_bundle_instance():
    """Copying a BundleInstance"""
    # This generally fails when run in the debugger, but seems alright stand-alone (?)
    copy.copy(h.BundleInstance(name="inst", of=h.Bundle()))


def test_bundle_destructure():
    """Test de-structuring bundles to individual Signals"""

    @h.bundle
    class B:
        w1 = h.Signal(width=1)
        w3 = h.Signal(width=3)

    @h.module
    class Child:
        w1 = h.Port(width=1)
        w3 = h.Port(width=3)

    @h.module
    class Parent:
        b = B()  # `B`-type Bundle instance
        c = Child(w1=b.w1, w3=b.w3)  # `Child`-type instance, connected to `b`

    h.elaborate(Parent)

    assert len(Parent.instances) == 1
    assert len(Parent.signals) == 2
    assert "b_w1" in Parent.signals
    assert "b_w3" in Parent.signals
    assert Parent.c.conns["w1"] is Parent.signals["b_w1"]
    assert Parent.c.conns["w3"] is Parent.signals["b_w3"]


def test_sub_bundle_conn():
    """Test connecting via BundleRef to a sub-Bundle"""

    @h.bundle
    class B1:
        s = h.Signal()

    @h.bundle
    class B2:
        b1 = B1()

    @h.module
    class HasB1:
        b1 = B1(port=True)

    @h.module
    class HasB2:
        b2 = B2()
        hasb1 = HasB1(b1=b2.b1)

    h.elaborate(HasB2)


def test_nested_bundle_conn():
    """Test connecting to a nested bundle ref"""

    @h.bundle
    class B1:
        s = h.Signal()

    @h.bundle
    class B2:
        b1 = B1()

    @h.bundle
    class B3:
        b2 = B2()

    @h.bundle
    class B4:
        b3 = B3()

    @h.module
    class HasB1:
        b1 = B1(port=True)

    @h.module
    class HasB4:
        b4 = B4()
        hasb1 = HasB1(b1=b4.b3.b2.b1)

    h.elaborate(HasB4)


def test_nested_bundle_conn2():
    """Test some bundles which have multiple instances of the same bundle as attributes."""

    @h.bundle
    class Ab:
        a, b = h.Signals(2)

    @h.bundle
    class FourAbs:
        i0 = Ab()
        i1 = Ab()
        i2 = Ab()
        i3 = Ab()

    @h.module
    class HasAbPort:
        ab = Ab(port=True)

    @h.module
    class HasThose:
        four_abs = FourAbs()
        i0 = HasAbPort(ab=four_abs.i0)
        i1 = HasAbPort(ab=four_abs.i1)
        i2 = HasAbPort(ab=four_abs.i2)
        i3 = HasAbPort(ab=four_abs.i3)

    h.elaborate(HasThose)

    assert HasThose.instances["i0"].conns["ab_a"].name == "four_abs_i0_a"
    assert HasThose.instances["i0"].conns["ab_b"].name == "four_abs_i0_b"
    assert HasThose.instances["i1"].conns["ab_a"].name == "four_abs_i1_a"
    assert HasThose.instances["i1"].conns["ab_b"].name == "four_abs_i1_b"
    assert HasThose.instances["i2"].conns["ab_a"].name == "four_abs_i2_a"
    assert HasThose.instances["i2"].conns["ab_b"].name == "four_abs_i2_b"
    assert HasThose.instances["i3"].conns["ab_a"].name == "four_abs_i3_a"
    assert HasThose.instances["i3"].conns["ab_b"].name == "four_abs_i3_b"


def test_anon_bundle_port_conn():
    """Test connecting via PortRef to an AnonymousBundle"""

    @h.bundle
    class B:
        s = h.Signal()

    @h.module
    class HasB:
        b = B(port=True)

    @h.module
    class Top:
        s = h.Signal()
        h1 = HasB(b=h.AnonymousBundle(s=s))
        h2 = HasB(b=h1.b)
        h3 = HasB(b=h2.b)

    # Elaborate to flesh this out
    h.elaborate(Top)

    # Check this resolved to a single Signal in Top
    assert len(Top.signals) == 1
    assert len(Top.bundles) == 0
    assert len(Top.instances) == 3

    # And check that Signal is connected to all three Instances
    assert Top.h1.b_s is Top.s
    assert Top.h2.b_s is Top.s
    assert Top.h3.b_s is Top.s


def test_anon_bundle_refs():
    """Test adding `BundleRef`s to `AnonymousBundle`s."""

    @h.bundle
    class Diff:
        p, n = h.Signals(2)

    @h.module
    class HasDiff:
        # Module with a `Diff` Port
        d = Diff(port=True)

    @h.module
    class HasHasDiff:
        # Module instantiating a few `HasDiff`
        d = Diff()
        # Instance connected to the Bundle Instance
        h1 = HasDiff(d=d)

        # THE POINT OF THE TEST:
        # Instance with an Anon Bundle, "equal to" the Bundle Instance
        h2 = HasDiff(d=h.AnonymousBundle(p=d.p, n=d.n))
        # Instance with (p, n) flipped
        h3 = HasDiff(d=h.AnonymousBundle(p=d.n, n=d.p))

    h.elaborate(HasHasDiff)


@pytest.mark.xfail(reason="#68 https://github.com/dan-fritchman/Hdl21/issues/68")
def test_no_role_directions():
    """Test directions on Bundles without Roles"""

    @h.bundle
    class B:
        # A Bundle with directed Ports
        a = h.Input()
        b = h.Output()
        c = h.Inout()
        d = h.Port()
        e = h.Signal()

    @h.module
    class HasB:
        b = B(port=True)

    @h.module
    class HasHasB:
        b = B()
        h1 = HasB(b=b)

    h.elaborate(HasHasB)

    assert HasB.b_a.direction == h.PortDir.INPUT
    assert HasB.b_b.direction == h.PortDir.OUTPUT
    assert HasB.b_c.direction == h.PortDir.INPUT
    assert HasB.b_d.direction == h.PortDir.NONE
    assert HasB.b_e.direction == None


def test_re_elab_generator_with_bundle_portref():
    """Test re-elaborating a generator with a bundle port, and a `PortRef` to it.
    This can be problematic as bundle ports are flattened during elaboration."""

    @h.generator
    def G(_: h.HasNoParams) -> h.Module:
        @h.module
        class G:
            p = h.Diff(port=True)

        return G

    @h.generator
    def T(_: h.HasNoParams) -> h.Module:
        @h.module
        class T:
            g1 = G()()
            g2 = G()(p=g1.p)

        return T

    h.elaborate(T())
    h.elaborate(T())


def test_re_elab_bundle_port():
    """Test elaborating, then re-elaborating Modules with Bundle-ports."""

    @h.bundle
    class Ab:
        a, b = h.Signals(2)

    @h.module
    class Bot:
        ab = Ab(port=True)

    @h.module
    class Top1:
        ab = Ab()
        bot = Bot(ab=ab)

    # Elaborate `Top1`. This works just fine.
    h.elaborate(Top1)

    # Now the problem children
    @h.module
    class Top2:
        ab = Ab()
        bot = Bot(ab=ab)  # <= especially this here

    # Elaborating `Top2` fails, as `Bot`'s bundle-valued ports have been flattened
    h.elaborate(Top2)


def test_bundle_noconns():
    """Test connecting `NoConn` to bundle-valued ports"""

    @h.bundle
    class A:
        ...  # empty

    @h.bundle
    class B:
        s1 = h.Signal()

    @h.bundle
    class C:
        s1, s2, s3 = h.Signals(3)

    @h.module
    class Bot:
        a = A(port=True)
        b = B(port=True)
        c = C(port=True)

    @h.module
    class Top:
        bot1 = Bot(
            a=h.NoConn(),
            b=h.NoConn(),
            c=h.NoConn(),
        )

    # Add another instance procedurally, and connect it via PortRefs
    Top.bot2 = Bot()
    Top.bot2.a = h.NoConn()
    Top.bot2.b = h.NoConn()
    Top.bot2.c = h.NoConn()

    # Elaborate it, flesh all this out
    h.elaborate(Top)

    assert len(Bot.ports) == 4
    assert sorted(Bot.ports.keys()) == [
        "b_s1",
        "c_s1",
        "c_s2",
        "c_s3",
    ]
    assert len(Top.signals) == 8
    assert sorted(Top.signals.keys()) == [
        "bot1_b_s1",
        "bot1_c_s1",
        "bot1_c_s2",
        "bot1_c_s3",
        "bot2_b_s1",
        "bot2_c_s1",
        "bot2_c_s2",
        "bot2_c_s3",
    ]

    assert Top.bot1.conns["b_s1"] is Top.bot1_b_s1
    assert Top.bot1.conns["c_s1"] is Top.bot1_c_s1
    assert Top.bot1.conns["c_s2"] is Top.bot1_c_s2
    assert Top.bot1.conns["c_s3"] is Top.bot1_c_s3

    assert Top.bot2.conns["b_s1"] is Top.bot2_b_s1
    assert Top.bot2.conns["c_s1"] is Top.bot2_c_s1
    assert Top.bot2.conns["c_s2"] is Top.bot2_c_s2
    assert Top.bot2.conns["c_s3"] is Top.bot2_c_s3


def test_structured_roles():
    """Test creating structured Bundle Roles"""

    @h.bundle
    class MyBundle:
        Host, Device = 2 * h.Role()

    class HostDeviceEnum(Enum):
        Host = auto()
        Device = auto()

    HostDevice = h.RoleSet.from_enum(HostDeviceEnum)

    @h.module
    class HostModule:
        mb = MyBundle(role=HostDevice.Host, port=True)
        # FIXME: not this part, at least not yet
        # mb = MyBundle.Host()

    @h.module
    class DeviceModule:
        mb = MyBundle(role=HostDevice.Device, port=True)
        # FIXME: not this part, at least not yet
        # mb = MyBundle.Host()

    @h.module
    class Top:
        host = HostModule()
        device = DeviceModule(mb=host.mb)

    h.elaborate(Top)

    @h.bundle
    class AnotherBundle:
        roles = HostDevice

    @h.bundle
    class YetAnotherBundle:
        Roles = HostDevice  # Capital R also works

    assert AnotherBundle.roles is HostDevice
    assert AnotherBundle.Roles is HostDevice
    assert YetAnotherBundle.roles is HostDevice
    assert YetAnotherBundle.Roles is HostDevice


def test_flipped():
    """Test bundles with directed ports, and direction-flips thereof"""

    from hdl21.bundle import flippable

    @h.bundle
    class Empty:
        ...  # like it says, empty

    assert flippable(Empty)

    @h.bundle
    class NoFlip:
        s = h.Signal()

    assert not flippable(NoFlip)

    @h.bundle
    class B:
        i = h.Input()
        o = h.Output()
        io = h.Inout()
        p = h.Port()

    assert flippable(B)

    @h.module
    class M:
        b = B(port=True)

    @h.module
    class F:
        b = B(port=True, flipped=True)

    @h.module
    class Top:
        m = M()
        f = F(b=m.b)

    h.elaborate(Top)

    assert M.b_i.vis == h.Visibility.PORT
    assert M.b_i.direction == h.PortDir.INPUT
    assert M.b_o.vis == h.Visibility.PORT
    assert M.b_o.direction == h.PortDir.OUTPUT
    assert M.b_io.vis == h.Visibility.PORT
    assert M.b_io.direction == h.PortDir.INOUT
    assert M.b_p.vis == h.Visibility.PORT
    assert M.b_p.direction == h.PortDir.NONE

    assert F.b_i.vis == h.Visibility.PORT
    assert F.b_i.direction == h.PortDir.OUTPUT
    assert F.b_o.vis == h.Visibility.PORT
    assert F.b_o.direction == h.PortDir.INPUT
    assert F.b_io.vis == h.Visibility.PORT
    assert F.b_io.direction == h.PortDir.INOUT
    assert F.b_p.vis == h.Visibility.PORT
    assert F.b_p.direction == h.PortDir.NONE

    assert Top.m_b_i.vis == h.Visibility.INTERNAL
    assert Top.m_b_i.direction == h.PortDir.NONE
    assert Top.m_b_o.vis == h.Visibility.INTERNAL
    assert Top.m_b_o.direction == h.PortDir.NONE
    assert Top.m_b_io.vis == h.Visibility.INTERNAL
    assert Top.m_b_io.direction == h.PortDir.NONE
    assert Top.m_b_p.vis == h.Visibility.INTERNAL
    assert Top.m_b_p.direction == h.PortDir.NONE


def test_nested_flipping():
    """# Test nesting the `flipped` attribute"""

    @h.bundle
    class Base:
        i = h.Input()
        o = h.Output()

    @h.bundle
    class Nested:
        b = h.flipped(Base())
        ni = h.Input()
        no = h.Output()

    @h.bundle
    class NestedSquared:
        n = h.flipped(Nested())
        n2i = h.Input()
        n2o = h.Output()

    @h.module
    class M:
        n = h.flipped(NestedSquared(port=True))

    h.elaborate(M)

    # Some of these are a mouthful!
    # (That's the point of having nested Bundles really;
    # the flattened versions look pretty bad.)

    assert M.n_n2i.vis == h.Visibility.PORT
    assert M.n_n2i.direction == h.PortDir.OUTPUT
    assert M.n_n2o.vis == h.Visibility.PORT
    assert M.n_n2o.direction == h.PortDir.INPUT

    assert M.n_n_ni.vis == h.Visibility.PORT
    assert M.n_n_ni.direction == h.PortDir.INPUT
    assert M.n_n_no.vis == h.Visibility.PORT
    assert M.n_n_no.direction == h.PortDir.OUTPUT

    assert M.n_n_b_i.vis == h.Visibility.PORT
    assert M.n_n_b_i.direction == h.PortDir.OUTPUT
    assert M.n_n_b_o.vis == h.Visibility.PORT
    assert M.n_n_b_o.direction == h.PortDir.INPUT


def test_custom_elaborator():
    """Test creating a custom elaborator, and making it the global one"""

    from hdl21.elab import Elaborator, set_elaborator, reset_elaborator
    from hdl21.elab.passes import ResolvePortRefs

    e = Elaborator(passes=[ResolvePortRefs])
    set_elaborator(e)
    assert h.elab.the_global_elaborator is e
    reset_elaborator()
    assert h.elab.the_global_elaborator is not e



# Std-Lib Imports
from __future__ import annotations
from typing import Union
from decimal import Decimal

# Local Imports
from .datatype import _pydantic_major_version
from .prefix import Prefixed
from .literal import Literal


# The shared docstring
_doc = """
# The `Scalar` parameter type

Generally this means
```python
Union[Prefixed, Literal]
```
with built-in automatic conversions from each of:
```python
[str, int, float, Decimal]
```
when used in `paramclass` definitions.

`Scalar` is particularly designed for parameter-values of `Primitive`s and of simulations.
Most such parameters "want" to be the `Prefixed` type, for reasons outlined in
https://github.com/dan-fritchman/Hdl21#prefixed-numeric-parameters.

They often also need a string-valued escape hatch, e.g. when referring to out-of-Hdl21 quantities
such as parameters in external netlists or simulation decks.
These out-of-Hdl21 expressions are represented by the `Literal` type, a simple wrapper around `str`.

Where possible `Scalar` prefers to use the `Prefixed` variant.
Strings and built-in numbers (int, float, Decimal) are converted to `Prefixed` inline by the `validate` method.
All of the `validate` mechanisms work for `Scalar`s used as fields in `pydantic.dataclasses`.
which crucially include all `hdl21.paramclass`es.

While defined as a type, `Scalar` is not instantiable.
It is really a class-based statement of `Union[Prefixed, Literal]`, with class methods to aid in validation.

If writing "primitive-like" parameters - e.g. those that go into SPICE simulations,
or are provided to PDK-level devices, it is very likely that you will want to:

* Use `Scalar` as a parameter type, i.e. the `dtype` field of `Param`s.
* Never actually instantiate a `Scalar` directly, including for its default value.

Example:

```python
import hdl21 as h
from hdl21.prefix import NANO, µ
from decimal import Decimal

@h.paramclass
class MyMosParams:
w = h.Param(dtype=h.Scalar, desc="Width", default=1e-6) # Default `float` converts to a `Prefixed`
l = h.Param(dtype=h.Scalar, desc="Length", default="w/5") # Default `str` converts to a `Literal`

# Example instantiations
MyMosParams(w=Decimal(1e-6), l=3*µ)
MyMosParams(w=h.Literal("sim_param_width"), l=h.Prefixed.new(20, NANO))
MyMosParams(w="11*l", l=11)
```
"""


# Union of types convertible into `Scalar`
ToScalar = Union[Prefixed, Literal, str, int, float, Decimal]


def to_scalar(v: ToScalar) -> Union[Prefixed, Literal]:
    """# Validate and convert anything in the `ToScalar` set of types to a `Prefixed` or `Literal`.
    Most importantly this handles the case in which `v` is a *string*,
    which attempts conversion to a `Prefixed`, and falls back to a `Literal` on failure.
    """

    if isinstance(v, (Prefixed, Literal)):
        return v  # Valid as-is, return it.

    # Now the important case: strings
    if isinstance(v, str):
        try:  # Try to convert to a Prefixed, which internally converts to a Decimal
            return Prefixed(number=v)
        except Exception:  # Catch all exceptions
            return Literal(text=v)

    # Everything else - notably including `int` and `float` - must be convertible to `Prefixed`, or fails in its validation.
    return Prefixed(number=v)


if _pydantic_major_version == 1:
    from .datatype import BaseModel

    class Scalar(BaseModel):
        # The Pydantic "custom root types" feature is really what makes this work:
        # https://docs.pydantic.dev/latest/usage/models/#custom-root-types
        __root__: Union[Prefixed, Literal]
        __doc__ = _doc

        def __init__(self, *_, **__):
            # Brick any attempts to create instances
            msg = f"Invalid attempt to instantiate a `Scalar` directly. "
            msg += f"Create either of its variants `Prefixed` or `Literal` instead, "
            msg += f"or use their built-in conversions from strings, ints, floats, and Decimals."
            raise RuntimeError(msg)

        @classmethod
        def __get_validators__(cls):
            yield to_scalar

else:
    from .datatype import BeforeValidator
    from typing import Annotated

    # Union of types convertible into `Scalar`
    Scalar = Annotated[
        Union[Prefixed, Literal],
        BeforeValidator(to_scalar),
    ]

__all__ = ["Scalar", "ToScalar", "to_scalar"]
__doc__ = _doc



"""
# PDK Corners
"""

from enum import Enum, auto
from pydantic.dataclasses import dataclass


class Corner(Enum):
    """# Enumerated Generic Corners

    Values of the `Corner` enum can mean either the variations in a particular device,
    e.g. the "slow" versus "fast" variations of a poly resistor,
    or can just as often refer to a set of such variations within a given technology.

    In the latter case `Corner` values are often expanded by PDK-level code to include each
    constituent device variation. For example `my.pdk.corner(Corner.FAST)` might expand to
    definitions of "fast" Cmos transistors, resistors, and capacitors."""

    TYP = auto()
    FAST = auto()
    SLOW = auto()

    def __repr__(self) -> str:
        return self.name


class CornerType(Enum):
    """Enumerated Types of PDK Content to be Corner-Varied"""

    MOS = auto()  # A single transistor type, e.g. NMOS
    CMOS = auto()  # A pair of transistors, e.g. NMOS/PMOS
    RES = auto()  # Resistors
    CAP = auto()  # Capacitors

    def __repr__(self) -> str:
        return self.name


@dataclass
class CmosCornerPair:
    """Cmos Corner Pair"""

    nmos: Corner
    pmos: Corner


class CmosCorner(Enum):
    """# Cmos Corner
    All values are `CmosCornerPair` objects.
    Naming convention is (Nmos, Pmos), e.g. "FS" is fast-N, slow-P."""

    TT = CmosCornerPair(Corner.TYP, Corner.TYP)
    FF = CmosCornerPair(Corner.FAST, Corner.FAST)
    SS = CmosCornerPair(Corner.SLOW, Corner.SLOW)
    FS = CmosCornerPair(Corner.FAST, Corner.SLOW)
    SF = CmosCornerPair(Corner.SLOW, Corner.FAST)

    def __repr__(self) -> str:
        return self.name
